(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

; --------------- Basic Definitions -------------------

(declare-sort Obj 0)

(declare-const Null Obj)
(declare-const Uninitialized Obj)
(assert (! (distinct Null Uninitialized) :named Distinct_Null_Uninitialized))

; --------------- Classes -------------------

(declare-sort Class 0)

(declare-fun class(Obj) Class)
(declare-const class_object Class)
(declare-const class_nullclass Class)
(declare-const class_boolean Class)
(declare-const class_numeric Class)
(declare-const class_listItem Class)
(declare-const class_list Class)
(declare-const class_map_item Class)
(declare-const class_map Class)
(declare-const class_main Class)
(assert (! (distinct class_object class_nullclass class_boolean class_numeric class_listItem class_list class_map_item class_map class_main) :named classes_distinct))
(assert (! (= (class Null) class_nullclass) :named Null_class))

; --------------- Instance variables -------------------

(declare-sort IVar 0)

(declare-const iv_as_ref IVar)
(declare-const iv_pred IVar)
(declare-const iv_next IVar)
(declare-const iv_value IVar)
(declare-const iv_first IVar)
(declare-const iv_last IVar)
(declare-const iv_key IVar)
(declare-const iv_map IVar)
(assert (! (distinct iv_as_ref iv_pred iv_next iv_value iv_first iv_last iv_key iv_map) :named ivar_distinct ))

(declare-fun inst_ref(Obj IVar) Obj)
(declare-const this Obj)

; Null must not reference anything
(assert (! (forall ((iv IVar)) (= (inst_ref Null iv) Uninitialized)) :named Null_references_nothing))

; --------------- Predicates -------------------------------


; from lib/stdlib.theory.smt:
; bounds
(declare-const num_bound Int)
(declare-const list_bound Int)
(assert (= num_bound 5))
(assert (= list_bound 5))

; bool

(define-fun is_B((o Obj)) Bool (= (class o) class_boolean))

(define-fun map_B((o Obj) (b Bool)) Bool (and
                                         (is_B o)
                                         (= b (not (= (inst_ref o iv_as_ref) Null)))))


; numeric

(define-fun is_N((o Obj)) Bool (= (class o) class_numeric))

; recursive definition
;(define-fun map_N((o Obj) (n Int)) Bool (and
;                                   (is_N o)
;                                   (or
;                                     (and (= (inst_ref o iv_pred) Null) (= n 0))
;                                     (map_N (inst_ref o iv_pred) (- n 1)))))

; quantified definition with bound
;(define-fun map_N((o Obj) (n Int)) Bool
;                              (and
;                                (<= 0 n)
;                                (< n num_bound)
;				(exists ((a_o (Array Int Obj)))
;				  (and
;                                    (>= n 0)
;			            (forall ((i Int))
;				      (=> (and (<= 0 i) (< i n))
;				        (and
;				          (is_N (select a_o i))
;				          (= (inst_ref (select a_o i) iv_pred) (select a_o (+ i 1)))
;				        )
;				      )
;				    )
;				    (= (select a_o 0) o)
;				    (= (inst_ref (select a_o n) iv_pred) Null)
;				  ))))

; faking the mapping using an uninterpreted function.
(declare-fun map_N_internal (Obj) Int)

(define-fun map_N((o Obj) (n Int)) Bool
             (and
               (= (class o) class_numeric)
               (= (map_N_internal o) n)
             )
)

; build-in type List (Tuple of size and contents)

(declare-datatypes () ((MyList (mk-pair (size Int) (contents (Array Int Obj))))))

(define-fun size_list ((l MyList) (size Int)) Bool
                      (and
                        (<= 0 size)
                        (< size list_bound)
                        (= (size l) size)
                      )
)

(define-fun select_list ((l MyList) (index Int) (result Obj)) Bool
                        (and
                          (exists ((ll Int))
                            (and
                              (size_list l ll)
                              (>= index 0)
                              (< index ll)
                            )
                          )
                          (= (select (contents l) index) result)
                        )
)

(define-fun in_list ((o Obj) (l MyList)) Bool
                    (exists ((i Int)) (select_list l i o)))


; list (objects)

(define-fun is_L ((l Obj)) Bool (= (class l) class_list))

;list (with the mapping faked by an uninterpreted function)

(declare-fun map_L_internal ((Obj)) MyList)

(define-fun map_L ((o Obj) (l MyList)) Bool
                  (and
                    (is_L o)
                    (= l (map_L_internal o))
                  ))

(define-fun select_L ((o Obj) (index Int) (result Obj)) Bool
                     (and
                       (is_L o)
                       (select_list (map_L_internal o) index result)
                     ))

(define-fun size_L ((o Obj) (size Int)) Bool
                     (and
                       (is_L o)
                       (size_list (map_L_internal o) size)
                     ))

(define-fun in_L ((o Obj) (l Obj)) Bool
                 (and
                   (is_L o)
                   (in_list o (map_L_internal o))
                 )
)

; maps

(define-sort MyMap () (Array Obj Obj))

(define-fun maps_to ((m MyMap) (k Obj) (v Obj)) Bool
                    (= (select m k) v)
)

; ================== Property-Specific ======================

(declare-const const_value Obj)
(declare-const const_var Obj)
(declare-const const_op Obj)
(declare-const const_add Obj)
(assert (! (distinct const_value const_var const_op const_add Null) :named distinct_const_objects))

;def valuetree2(t:list,n:num) = map_N(t[0],VALUE) && map_N(t[1],n);
(define-fun valuetree2 ((t MyList) (n Int)) Bool
                   (and
                     (exists ((o2 Obj))
                       (and
                         (select_list t 0 const_value)
                         (select_list t 1 o2)
                         (map_N o2 n)
                       )
                     ) 
                   )
)

;def valuetree1(t:list) = exists n:num. valuetree2(t,n);
(define-fun valuetree1 ((t MyList)) Bool
                   (exists ((n Int))
                     (valuetree2 t n)
                   )
)

; def vartree3(t:list,e:map,x:string) = map_N(t[0],VAR) && mapstring(t[1],x) && e[x] != null
(define-fun vartree3 ((t MyList) (e MyMap) (x Obj)) Bool
                   (and
                     (select_list t 0 const_var)
                     (select_list t 1 x)
                     (exists ((o Obj))
                       (and
                         (maps_to e x o)
                         (is_N o)
                       )
                     )
                   )
)

;def vartree2(t:list,e:map) = exists x:obj. vartree3(t,e,x);
(define-fun vartree2 ((t MyList) (e MyMap)) Bool
                   (exists ((x Obj))
                     (vartree3 t e x)
                   )
)

;def optree5(t:list,e:map,op:num,l:list,r:list) = map_N(t[0],OP) && map_N(t[1],op) && map_L(t[2],l) && parsetree2(l,e) && map_L(t[3],r) && parsetree2(r,e)
;(define-fun optree5 ((t MyList) (e MyMap) (op Int) (l MyList) (r MyList)) Bool
;                 (exists ((o1 Obj) (o2 Obj) (o3 Obj))
;                   (and
;                     (select_list t 0 const_op)
;                     (select_list t 1 o1)
;                     (map_N o1 op)
;                     (select_list t 2 o2)
;                     (map_L o2 l)
;                     (parsetree2 l e)
;                     (select_list t 3 o3)
;                     (map_L o3 r)
;                     (parsetree2 r e)
;                   )
;                 )
;)


;def parsetree2(t:list,e:map) = valuetree1(t) || vartree2(t,e) || optree2(t,e);
;(define-fun parsetree2 ((t MyList) (e MyMap)) Bool
;                   (or
;                     (valuetree1 t)
;                     (vartree2 t e)
;                     (optree2 t e)
;                   )
;)

;(define-fun parsetree2 ((t MyList) (e MyMap)) Bool
;                   (or
;                     (valuetree1 t)
;                     (vartree2 t e)
;                     (exists ((op Int) (l MyList) (r MyList))
;                       (and
;                         (map_N (select t 0) const_op)
;                         (map_N (select t 1) op)
;                         (map_L (select t 2) l)
;                         (parsetree2 l e)
;                         (map_L (select t 3) r)
;                         (parsetree2 r e)
;                       )
;                     )
;                   )
;)

; --------------- Tree Recursion -------------------

(declare-const MAX_LEVEL Int)
(assert (= MAX_LEVEL 4))

(define-fun pow2 ((n Int)) Int
                    (ite (= n 0) 1
                    (ite (= n 1) 2
                    (ite (= n 2) 4
                    (ite (= n 3) 8
                    (ite (= n 4) 16
                    32
                    ))))))

(define-fun s-level ((l Int)) Int
                    (- (pow2 l) 1)
                    )

(define-fun size-level ((l Int)) Int
                    (pow2 l)
                    )

(define-fun index-of ((level Int) (index Int)) Int
                         (+ (s-level level) index)
                         )

(define-fun child-a ((level Int) (index Int) (level-child Int) (index-child Int)) Bool
                    (and (= level-child (+ level 1))
                         (= index-child (* index 2))
                         ))

(define-fun child-b ((level Int) (index Int) (level-child Int) (index-child Int)) Bool
                    (and (= level-child (+ level 1))
                         (= index-child (+ (* index 2) 1))
                         ))

;(define-fun optree ((t MyList) (l MyList) (r MyList)) Bool
;                               (exists ((op Int) (o1 Obj) (o2 Obj) (o3 Obj))
;                                 (and
;                                   (select_list t 0 const_op)
;                                   (select_list t 1 o1)
;                                   (map_N o1 op)
;                                   (select_list t 2 o2)
;                                   (map_L o2 l)
;                                   (select_list t 3 o3)
;                                   (map_L o3 r)
;                                 )
;                               )
;)

;(define-fun optree1 ((t MyList)) Bool
;                          (exists ((l MyList) (r MyList))
;                            (optree t l r)
;                          )
;)

(define-fun optree ((t MyList) (l Obj) (r Obj)) Bool
                               (exists ((op Int) (o1 Obj) (o2 Obj) (o3 Obj))
                                 (and
                                   (select_list t 0 const_op)
                                   (select_list t 1 const_add)
                                   (select_list t 2 l)
                                   (select_list t 3 r)
                                 )
                               )
)

(define-fun optree1 ((t MyList)) Bool
                          (exists ((l Obj) (r Obj))
                            (optree t l r)
                          )
)

(define-fun parsetree2 ((t MyList) (e MyMap)) Bool
                   (exists ((a (Array Int MyList)) (o (Array Int Obj)))
                     (and
                       (= (select a (index-of 0 0)) t)
                       (forall ((i Int) (li Int))
                        (and
                         (=>
                           (and (<= 0 li) (< li MAX_LEVEL) (<= 0 i) (< i (size-level li)))
                           (or
                             (valuetree1 (select a (index-of li i)))
                             (vartree2 (select a (index-of li i)) e)
                             (forall ((lj Int) (j1 Int) (j2 Int))
                               (=>
                                 (and (child-a li i lj j1) (child-b li i lj j2))
                                 (optree (select a (index-of li i))
                                         (select o (index-of lj j1))
                                         (select o (index-of lj j2))
                                 )
                               )
                             )
                           )
                         )
                         (map_L (select o (index-of li i)) (select a (index-of li i)))
                        )
                       )
                     )
                   )
)

(define-fun optree5 ((t MyList) (e MyMap) (op Int) (l MyList) (r MyList)) Bool
                  (and
                    (parsetree2 t e)
                    (not (valuetree1 t))
                    (not (vartree2 t e))
                  )
)

(declare-fun parsetree-uf (MyList MyMap) Bool)

(define-fun optree-uf ((t MyList) (e MyMap)) Bool
                      (and
                        (select_list t 0 const_op)
                        (select_list t 1 const_add)
                        (exists ((l Obj) (r Obj) (ll MyList) (rl MyList))
                          (and
                            (select_list t 2 l)
                            (map_L l ll)
                            (select_list t 3 r)
                            (map_L r rl)
                            (parsetree-uf ll e)
                            (parsetree-uf rl e)
                          )
                        )
                      )
)

; def optree2(t:list,e:map) = exists op:num. exists l:list. exists r:list. optree5(t,e,op,l,r);
(define-fun optree2 ((t MyList) (e MyMap)) Bool
                   (exists ((op Int) (l MyList) (r MyList))
                     (optree5 t e op l r)
                   )
)

(declare-fun treeval-uf (MyList MyMap Int) Bool)
(declare-fun plus-uf (Int Int) Int)

; def treeval3(tree:list,env:map,n:num) = valuetree2(tree,n) || exists x:string. vartree3(tree,env,x) && env[x] = n || exists op,n_l,n_r:num,l,r:list. optree5(tree,env,op,l,r) && treeval3(l,env,n_l) && treeval3(r,env,n_r) && op = ADD implies n = n_l + n_r ...
(define-fun treeval-unroll ((t MyList) (e MyMap) (n Int)) Bool
                   (or
                     (valuetree2 t n)
                     (exists ((x Obj) (v Obj))
                       (and
                         (vartree3 t e x)
                         (maps_to e x v)
                         (map_N v n)
                       )
                     )
                     (exists ((nl Int) (nr Int) (ol Obj) (or Obj) (l MyList) (r MyList))
                       (and
                         (optree t ol or)
                         (map_L ol l)
                         (map_L or r)
                         (treeval-uf l e nl)
                         (treeval-uf r e nr)
                         (= n (plus-uf nl nr))
                       )
                     )
                   )
)

;(define-fun treeval3 ((t MyList) (e MyMap) (n Int)) Bool
;                     (exists ((a (Array Int MyList)) (res (Array Int Int)))
;                       (and
;                         (= (select a (index-of 0 0)) t)
;                         (= (select res (index-of 0 0)) n)
;                       )
;                     )
;)

; =============== Testing ===============0

;#invariant ${\color{blue}\forall t:\typeALlist,e:\typeALmap \quantsep \mathit{parsetree2(t,e)} \rightarrow}$
;#       ${\color{blue}t[0] = \ProgVirtual{\PL{VALUE}} \rightarrow \mathit{valuetree1(t)} \wedge t[0] =
;# \ProgVirtual{\PL{VAR}} \rightarrow \mathit{vartree2(t,e)} \wedge t[0] = \ProgVirtual{\PL{OP}} \rightarrow
;# \mathit{optree2(t,e)}}$;

(push)

(declare-const tree MyList)
(declare-const env MyMap)

(echo "Functional Correctness Verification Condition (VC1)")

; the invariant can be deduced from the definitions of the predicates
(assert (not
  (=>
    (and
      (parsetree2 tree env)
      (select_list tree 0 const_value)
    )
    (valuetree1 tree)
  )))


(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC2)")

(declare-const tree MyList)
(declare-const env MyMap)

; the invariant can be deduced from the definitions of the predicates
(assert (not
  (=>
    (and
      (parsetree2 tree env)
      (select_list tree 0 const_var)
    )
    (vartree2 tree env)
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC3)")

(declare-const tree MyList)
(declare-const env MyMap)

; the invariant can be deduced from the definitions of the predicates
(assert (not
  (=>
    (and
      (parsetree2 tree env)
      (select_list tree 0 const_op)
    )
    (optree1 tree)
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC4)")
(echo "[treeval unrolled once, all occurrences of treeval and + replaced by UF]")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (valuetree1 tree)
    (exists ((o Obj) (n Int))
      (and
        (select_list tree 1 o)
        (map_N o n) 
        (treeval-unroll tree env n)
      )
    )
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC5)")
(echo "[treeval unrolled once, all occurrences of treeval and + replaced by UF]")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (vartree2 tree env)
    (exists ((x Obj) (o Obj) (n Int))
      (and
        (select_list tree 1 x)
        (maps_to env x o)
        (map_N o n)
        (treeval-unroll tree env n)
      )
    )
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC6)")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (optree-uf tree env)
    (exists ((o Obj) (t2 MyList))
      (and
        (select_list tree 2 o)
        (map_L o t2)
        (parsetree-uf t2 env)
      )
    )
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC7)")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (optree-uf tree env)
    (exists ((o Obj) (t2 MyList))
      (and
        (select_list tree 3 o)
        (map_L o t2)
        (parsetree-uf t2 env)
      )
    )
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC8)")
(echo "[unrolled treeval once (on the right-hand side), replaced all occurrences of treeval and + by UF]")
(echo "[split into 2 parts - part1]")

(declare-const tree MyList)
(declare-const env MyMap)
(declare-const nl Int)
(declare-const nr Int)

; original VC8
;(assert (not
;  (=>
;    (exists ((o2 Obj) (o3 Obj) (l2 MyList) (l3 MyList))
;      (and
;        (select_list tree 0 const_op)
;        (select_list tree 1 const_add)
;        (select_list tree 2 o2)
;        (map_L o2 l2)
;        (select_list tree 3 o3)
;        (map_L o3 l3)
;        (treeval-uf l2 env nl)
;        (treeval-uf l3 env nr)
;      )
;    )
;    (treeval-unroll tree env (plus-uf nl nr))
;  )))

(assert (not
  (=>
    (exists ((o2 Obj) (o3 Obj) (l2 MyList) (l3 MyList))
      (and
        (select_list tree 0 const_op)
        (select_list tree 1 const_add)
        (select_list tree 2 o2)
        (map_L o2 l2)
        (select_list tree 3 o3)
        (map_L o3 l3)
        (treeval-uf l2 env nl)
        (treeval-uf l3 env nr)
      )
    )
    (exists ((nl1 Int) (nr1 Int) (ol Obj) (or Obj) (l MyList) (r MyList))
         (and
           (optree tree ol or)
           (map_L ol l)
           (map_L or r)
           (treeval-uf l env nl1)
           (treeval-uf r env nr1)
           (= (plus-uf nl nr) (plus-uf nl1 nr1))
         )
    )
  )))



(check-sat)

(pop)
(push)

(echo "[split into 2 parts - part2]")

(declare-const tree MyList)
(declare-const env MyMap)
(declare-const nl Int)
(declare-const nr Int)

;(assert (not
;  (=>
;    (exists ((o2 Obj) (o3 Obj) (l2 MyList) (l3 MyList))
;      (and
;        (select_list tree 0 const_op)
;        (select_list tree 1 const_add)
;        (select_list tree 2 o2)
;        (map_L o2 l2)
;        (select_list tree 3 o3)
;        (map_L o3 l3)
;        (treeval-uf l2 env nl)
;        (treeval-uf l3 env nr)
;      )
;    )
;    (treeval-unroll tree env (plus-uf nl nr))
;  )))

(assert (not
  (=>
    (exists ((nl1 Int) (nr1 Int) (ol Obj) (or Obj) (l MyList) (r MyList))
         (and
           (optree tree ol or)
           (map_L ol l)
           (map_L or r)
           (treeval-uf l env nl1)
           (treeval-uf r env nr1)
           (= (plus-uf nl nr) (plus-uf nl1 nr1))
         )
    )
    (treeval-unroll tree env (plus-uf nl nr))
  )))

(check-sat)

(echo "[(VC8) follows by transitivity of implication]")

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC9)")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (and
      (parsetree2 tree env)
      (not (select_list tree 0 const_value))
      (not (select_list tree 0 const_var))
      (not (select_list tree 0 const_op))
    )
    false
  )))

(check-sat)

(pop)
(push)

(echo "Functional Correctness Verification Condition (VC10)")

(declare-const tree MyList)
(declare-const env MyMap)

(assert (not
  (=>
    (and
      (parsetree2 tree env)
      (select_list tree 0 const_op)
      (not (select_list tree 1 const_add))
    )
    false
  )))

(check-sat)

(pop)
(push)

