(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

(push)

(echo "VC1 with Layer of Abstraction")

(declare-const pv_x Int)
(declare-const pv_y Int)
(declare-const lv_x Int)
(declare-const lv_y Int)

(assert (not 
(=> 
  (and (= pv_x lv_x) (= pv_y lv_y))
  (and
    (= lv_x (- (+ pv_x pv_y) pv_y))
    (= lv_y (- (+ pv_x pv_y) (- (+ pv_x pv_y) pv_y)))    
  )
)))


(check-sat)

(pop)
(push)

(echo "VC1 Type Independent")

(declare-sort X)

(declare-fun plus (X X) X)
(declare-fun minus (X X) X)
(declare-fun neg (X) X)

(declare-const zero X)

; Commutativity
(assert (forall ((a X) (b X))
          (= (plus a b) (plus b a))
        )
)

; Associativity
(assert (forall ((a X) (b X) (c X))
          (= (plus (plus a b) c) (plus a (plus b c)))
        )
)

; Neutral Element (right)
(assert (forall ((a X) (b X) (c X))
          (= (plus a zero) a)
        )
)

; Neutral Element (left)
(assert (forall ((a X) (b X) (c X))
          (= (plus zero a) a)
        )
)

; Inverse Elements
(assert (forall ((a X))
          (= (plus a (neg a)) zero)
        )
)

; definition of minus
(assert (forall ((a X) (b X))
          (= (minus a b) (plus a (neg b)))
        )
)


(declare-const pv_x X)
(declare-const pv_y X)
(declare-const lv_x X)
(declare-const lv_y X)

(assert (not
  (=>
    (and (= pv_x lv_x) (= pv_y lv_y))
    (and
      (= lv_x (minus (plus pv_x pv_y) pv_y))
      (= lv_y (minus (plus pv_x pv_y) (minus (plus pv_x pv_y) pv_y)))
    )
  )
))

(check-sat)

