(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

; --------------- Basic Definitions -------------------

(declare-sort Obj 0)

(declare-const Null Obj)

; --------------- Classes -------------------

(declare-sort Class 0)

(declare-fun class(Obj) Class)
(declare-const class_object Class)
(declare-const class_nullclass Class)
(declare-const class_boolean Class)
(declare-const class_numeric Class)
(declare-const class_list Class)
(declare-const class_main Class)
(assert (distinct class_object class_nullclass class_boolean class_numeric class_list class_main))
(assert (= (class Null) class_nullclass))

; --------------- Instance variables -------------------

(declare-sort IVar 0)

(declare-const iv_as_ref IVar)
(declare-const iv_pred IVar)
(declare-const iv_next IVar)
(declare-const iv_value IVar)
(declare-const iv_size IVar)
(assert (distinct iv_as_ref iv_pred iv_next iv_value iv_size))

(declare-fun inst_ref(Obj IVar) Obj)
(declare-const this Obj)

; --------------- Predicates -------------------------------


; from lib/stdlib.theory.smt:
; bool

(define-fun is_B((o Obj)) Bool (= (class o) class_boolean))

(define-fun map_B((o Obj) (b Bool)) Bool (and
                                         (is_B o)
                                         (= b (not (= (inst_ref o iv_as_ref) Null)))))


; numeric

(define-fun is_N((o Obj)) Bool (= (class o) class_numeric))

; recursive definition
;(define-fun map_N((o Obj) (n Int)) Bool (and
;                                   (is_N o)
;                                   (or
;                                     (and (= (inst_ref o iv_pred) Null) (= n 0))
;                                     (map_N (inst_ref o iv_pred) (- n 1)))))

; quantified definition
(define-fun map_N((o Obj) (n Int)) Bool
				(exists ((a_o (Array Int Obj)))
				  (and
			            (forall ((i Int))
				      (=> (and (<= 0 i) (< i n))
				        (and
				          (is_N (select a_o i))
				          (= (inst_ref (select a_o i) iv_pred) (select a_o (+ i 1)))
				        )
				      )
				    )
				    (= (select a_o 0) o)
				    (= (inst_ref (select a_o n) iv_pred) Null)
				  )))

; list

;(define-fun is_L ((l Obj)) Bool (= (class l) class_list))

;(define-fun select_L ((l Obj) (i Int) (o Obj)) Bool
;                     (exists ((a (Array Int Obj)))
;                             (and (= (select a 0) l)
;                                  (forall ((i2 Int))
;                                          (=> (and (<= 0 i2) (< i2 i))
;                                              (= (inst_ref (select a i2) iv_next) (select a (+ i2 1)))))
;                                  (= (inst_ref (select a i) iv_value) o))))

;(define-fun size_L ((l Obj) (n Int)) Bool
;                   (exists ((o Obj)) (and (= (inst_ref l iv_size) o) (map_N o n))))

; predicate in_L(o:O, l:O*) = exists i:N. select_L(i,l,o)
;(define-fun in_L ((o Obj) (l Obj)) Bool
;                 (exists ((i Int)) (select_L l i o)))



; from ../../verification/thesis/material/case studies/in-place value switching/swap.theory.smt:


; --------------- Contextual Information -------------------

(declare-const lv_true Bool)
(declare-const lv_false Bool)

(assert (= lv_true true))
(assert (= lv_false false))

; TODO: local program variables


; --------------- Contextual Information -------------------
(declare-const pv_self Obj)
(assert (= (class pv_self) class_list))

(declare-const lv_s Int)

; --------------- Verification Condition -------------------
(declare-const pv_other Obj)

(declare-const pv_res Obj)
(declare-const pv_x Obj)
(declare-const lv_x Int)
(declare-const pv_y Obj)
(declare-const lv_y Int)

(assert (not
(=>
  (exists ((lv_x2 Int) (lv_y2 Int))
    (and
      (map_N pv_x lv_x2)
      (map_N pv_y lv_y2)
      (= lv_y (- lv_x2 lv_y2))
      (= lv_x lv_y2)
    )
  )
  (exists ((lv_s Int) (lv_o Int))   ; Rule of Adaptation for x = x - y
    (and
      (map_N pv_x lv_s)
      (map_N pv_y lv_o)
      (> lv_s lv_o)
      (forall ((pv_res Obj) (pv_s Obj) (pv_o Obj))
        (=>
          (map_N pv_res (- lv_s lv_o))
          (and
            (map_N pv_res lv_y)
            (map_N pv_y lv_x)
          )
        )
      )
    )
  )
)))


; --------------- Verify -------------------
(check-sat)
