#!/usr/bin/ruby

start_time = Time.now

# workaround to get treetop's require running on ruby 1.9
$LOAD_PATH << '.'

require 'rubygems'
#require 'ruby-debug'
require 'treetop'
require_relative 'dyn_parsing'
require_relative 'dyn-semantics'
require_relative 'dyn_type_inference'
require_relative 'ssa_transform'
require_relative 'serializer'

#TODO: parse & bootstrap STDLIB and program seperatly (so line numbers do not get skrewed up)
#TODO: exception-like stacktraces on errors

$stderr.puts Dir.pwd
Treetop.load("interpreter/dyn")
#require_relative 'dyn.treetop'

STDLIB = "lib/stdlib.dyn"

inference = false
execution = true
show_configuration = false
options = {}
while ARGV[0] =~ /^--/
  arg = ARGV.shift
  if arg == "--infer"
    inference = true
    options = {:output_solution => true}
  elsif arg == "--infer-verbose"
    inference = true
    options = {:output_solution => true, :output_constraints => true}
  elsif arg == "--typecheck"
    inference = true
  elsif arg == "--dontrun"
    execution = false
  elsif arg == "--show-configuration"
    show_configuration = true
  else
    raise RuntimeError.new("unknown argument #{arg.inspect}")
  end
end
filename = ARGV.shift
abort("no filename given") unless filename
prog = File.read(STDLIB) + File.read(filename)

lib = ParsedFile.new(STDLIB)
prog = ParsedFile.new(filename)

# parsing
parser = DynParser.new
lib.parse(parser)
prog.parse(parser)
parse_tree = lib.join_parse_trees(prog)

=begin
parser = DynParser.new
result = parser.parse(prog)
unless result
  puts parser.failure_reason
  puts parser.failure_line
  puts parser.failure_column
  exit(-1)
end
parse_tree = result.construct
=end

# type inference
if inference
    # transforming to static single-assignment form to make the analysis flow-sensitive
    begin
      ssa = SSATransformer.new()
      ssa_parse_tree = ssa.transform(parse_tree.dclone())
    rescue DynTypeError => e
      puts e.message()
      raise e
    end

    # DEBUG
    File.open("ssa-form.dyn","w") do |f|
      f.puts(Serializer.new.serialize(ssa_parse_tree))
    end

    begin
      inferencer = DynTypeInferencer.new(ssa_parse_tree)
    rescue BootstrapError => e
      puts "ERROR: #{e.message}"
      exit(-1)
    end
    begin
      type = inferencer.infer_types(options)
    rescue DynTypeError => e
      puts e.message()
      raise e
    rescue SystemStackError => e
      puts "Error: " + e.message()
      p e.backtrace
      raise e
    end
end

# execution
if execution
    begin
      configuration = DynConfiguration.new(parse_tree)
    rescue BootstrapError => e
      puts "ERROR: #{e.message}"
      exit(-1)
    end

    return_value = 0
    if configuration.eval()
      puts "finished"
      puts configuration.result()
      end_time = Time.now
      puts "needed #{(end_time - start_time).to_s} seconds"
    else
      puts configuration.error_text()
      return_value = -1
    end
    if show_configuration
      puts "--- FINAL CONFIGURATION ---"
      configuration.output_store()
    end
    exit(return_value)
end

