
#!/usr/bin/ruby

class Serializer

  def initialize()
    @visited = {}
  end

  def serialize(subtree, visited = {})
        if visited.has_key?(subtree.__id__)
          raise RuntimeError.new("loop detected! #{visited[subtree.__id__].inspect}")
        end
        visited = visited.merge(subtree.__id__ => subtree)

        case subtree[0]
          when :program
            return subtree[1..subtree.size()-2].map do |classdef|
              serialize(classdef, visited)
            end.join("\n\n")

	  when :classdef
            name = subtree[1]
            parent = subtree[2]
            methods = subtree[3]
            return "class #{name} < #{parent} {\n" + methods.map { |m| serialize(m,visited) }.join("\n\n") + "\n}\n"
	  when :rename
	    from = subtree[1]
	    to = subtree[2]
	    return "rename #{from} #{to}"

	  when :method
            name = subtree[1]
            params = subtree[2]
            body = subtree[3]
            return "method #{name}(#{params.join(',')}) {\n" + serialize(body, visited) + "\n}\n"

          when :variable
            var = subtree[1]
            var = "#{var[0]}_#{var[1].to_s}" if var.is_a?(Array)
            return var
  
          when :instance_variable
            var = subtree[1]
            var = "@#{var.to_s}"
            var = "@#{var[0]}_#{var[1].to_s}" if var.is_a?(Array)
            return var
  
          when :self
            return "self"

          when :assignment
            var = subtree[1]
            var = "#{var[0]}_#{var[1].to_s}" if var.is_a?(Array)
            exp = subtree[2]
            return "#{var} = #{serialize(exp, visited)}"
  
          when :instance_assignment
            var = subtree[1]
            var = "@#{var[0]}_#{var[1].to_s}" if var.is_a?(Array)
            exp = subtree[2]
            return "@#{var} = #{serialize(exp, visited)}"
  
          when :method_call
            recv = subtree[1]
            name = subtree[2]
            params = subtree[3]
	    return "#{serialize(recv, visited)}.#{name}(#{params.map{|p| serialize(p, visited)}.join(',')})"
            return currNode
  
          when :constructor_call
            classname = subtree[1]
            params = subtree[2]
            return "new #{classname}(#{params.map{|p| serialize(p, visited)}.join(',')})"
  
          when :call
            raise RuntimeError.new("did not expect call in parse tree to be serialized.")
  
          when :loc
            return "null" if subtree[1] == 0
            raise RuntimeError.new("did not expect loc in parse tree to be serialized: #{subtree.inspect}")
  
          when :if
            cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]
            return "if #{serialize(cond, visited)} then\n #{serialize(true_branch, visited)}\nelse\n#{serialize(false_branch, visited)}\nend\n"
            
          when :ifPS
	    cond = subtree[1]
            true_branch = subtree[2]
            false_branch = subtree[3]
            return "ifPS #{serialize(cond, visited)} then\n #{serialize(true_branch, visited)}\nelse\n#{serialize(false_branch, visited)}\nend\n"

          when :while
            cond = subtree[1]
            loop = subtree[2]
            return "while #{serialize(cond, visited)} do\n#{serialize(loop, visited)}\ndone\n"

          when :statement
            return subtree[1..subtree.size()-2].map { |stmt| serialize(stmt, visited) }.join(";\n")

          when :error
            return "error()"

          when :hook_loc_identical
            a = subtree[1]
            b = subtree[2]
            return "identical(#{serialize(a, visited)}, #{serialize(b, visited)})"

          when :hook_get_loc
            exp = subtree[1]
            return "get_loc(#{serialize(exp, visited)})"

          when :hook_is_a
            exp = subtree[1]
	    typestr = subtree[2]
            return "#{serialize(exp, visited)} is_a? #{typestr}"

          when :hook_psi
            return "psi(#{subtree[1..subtree.size()-2].map{|p| serialize(p, visited)}.join(",")})"

          when :hook_cast
            return "cast(#{subtree[1..subtree.size()-2].map{|p| serialize(p, visited)}.join(",")})"

	  when :uniontype
	    return "{#{subtree[1..subtree.size()-2].join(',')}}"

          when :assertion
            return "\#\{\?\!\?\}"

          else
            syserror("found unknown node in parsetree: #{subtree.inspect}")
        end
    
  end

  def syserror(text)
      raise RuntimeError.new("SYSERROR: " + text)
  end

end
