(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

; --------------- Basic Definitions -------------------

(declare-sort Obj 0)

(declare-const Null Obj)
(declare-const Uninitialized Obj)
(assert (! (distinct Null Uninitialized) :named Distinct_Null_Uninitialized))

; --------------- Classes -------------------

(declare-sort Class 0)

(declare-fun class(Obj) Class)
(declare-const class_object Class)
(declare-const class_nullclass Class)
(declare-const class_boolean Class)
(declare-const class_numeric Class)
(declare-const class_listItem Class)
(declare-const class_list Class)
(declare-const class_map_item Class)
(declare-const class_map Class)
(declare-const class_main Class)
(assert (! (distinct class_object class_nullclass class_boolean class_numeric class_listItem class_list class_map_item class_map class_main) :named classes_distinct))
(assert (! (= (class Null) class_nullclass) :named Null_class))

; --------------- Instance variables -------------------

(declare-sort IVar 0)

(declare-const iv_as_ref IVar)
(declare-const iv_pred IVar)
(declare-const iv_next IVar)
(declare-const iv_value IVar)
(declare-const iv_first IVar)
(declare-const iv_last IVar)
(declare-const iv_key IVar)
(declare-const iv_map IVar)
(assert (! (distinct iv_as_ref iv_pred iv_next iv_value iv_first iv_last iv_key iv_map) :named ivar_distinct ))

(declare-fun inst_ref(Obj IVar) Obj)
(declare-const this Obj)

; Null must not reference anything
(assert (! (forall ((iv IVar)) (= (inst_ref Null iv) Uninitialized)) :named Null_references_nothing))

; --------------- Predicates -------------------------------


; from lib/stdlib.theory.smt:
; bounds
(declare-const num_bound Int)
(declare-const list_bound Int)
(assert (= num_bound 5))
(assert (= list_bound 5))

; bool

(define-fun is_B((o Obj)) Bool (= (class o) class_boolean))

(define-fun map_B((o Obj) (b Bool)) Bool (and
                                         (is_B o)
                                         (= b (not (= (inst_ref o iv_as_ref) Null)))))


; numeric

(define-fun is_N((o Obj)) Bool (= (class o) class_numeric))

; recursive definition
;(define-fun map_N((o Obj) (n Int)) Bool (and
;                                   (is_N o)
;                                   (or
;                                     (and (= (inst_ref o iv_pred) Null) (= n 0))
;                                     (map_N (inst_ref o iv_pred) (- n 1)))))

; quantified definition with bound
;(define-fun map_N((o Obj) (n Int)) Bool
;                              (and
;                                (<= 0 n)
;                                (< n num_bound)
;				(exists ((a_o (Array Int Obj)))
;				  (and
;                                    (>= n 0)
;			            (forall ((i Int))
;				      (=> (and (<= 0 i) (< i n))
;				        (and
;				          (is_N (select a_o i))
;				          (= (inst_ref (select a_o i) iv_pred) (select a_o (+ i 1)))
;				        )
;				      )
;				    )
;				    (= (select a_o 0) o)
;				    (= (inst_ref (select a_o n) iv_pred) Null)
;				  ))))

; faking the mapping using an uninterpreted function.
(declare-fun map_N_internal (Obj) Int)

(define-fun map_N((o Obj) (n Int)) Bool
             (and
               (= (class o) class_numeric)
               (= (map_N_internal o) n)
             )
)

; build-in type List (Tuple of size and contents)

(declare-datatypes () ((MyList (mk-pair (size Int) (contents (Array Int Obj))))))

(define-fun size_list ((l MyList) (size Int)) Bool
                      (and
                        (<= 0 size)
                        (< size list_bound)
                        (= (size l) size)
                      )
)

(define-fun select_list ((l MyList) (index Int) (result Obj)) Bool
                        (and
                          (exists ((ll Int))
                            (and
                              (size_list l ll)
                              (>= index 0)
                              (< index ll)
                            )
                          )
                          (= (select (contents l) index) result)
                        )
)

(define-fun in_list ((o Obj) (l MyList)) Bool
                    (exists ((i Int)) (select_list l i o)))


; list (objects)

(define-fun is_L ((l Obj)) Bool (= (class l) class_list))

;list (with the mapping faked by an uninterpreted function)

(declare-fun map_L_internal ((Obj)) MyList)

(define-fun map_L ((o Obj) (l MyList)) Bool
                  (and
                    (is_L o)
                    (= l (map_L_internal o))
                  ))


; ================== Property-Specific ======================

(declare-const const_value Obj)
(declare-const const_var Obj)
(declare-const const_op Obj)
(declare-const const_add Obj)
(assert (! (distinct const_value const_var const_op const_add Null) :named distinct_const_objects))


; maps

(define-sort MyMap () (Array Obj Obj))

(define-fun maps_to ((m MyMap) (k Obj) (v Obj)) Bool
                    (= (select m k) v)
)

; tree recursion

(declare-const MAX_LEVEL Int)
(assert (= MAX_LEVEL 4))

(define-fun pow2 ((n Int)) Int
                    (ite (= n 0) 1
                    (ite (= n 1) 2
                    (ite (= n 2) 4
                    (ite (= n 3) 8
                    (ite (= n 4) 16
                    32
                    ))))))

(define-fun s-level ((l Int)) Int
                    (- (pow2 l) 1)
                    )

(define-fun size-level ((l Int)) Int
                    (pow2 l)
                    )

(define-fun index-of ((level Int) (index Int)) Int
                         (+ (s-level level) index)
                         )

(define-fun child-a ((level Int) (index Int) (level-child Int) (index-child Int)) Bool
                    (and (= level-child (+ level 1))
                         (= index-child (* index 2))
                         ))

(define-fun child-b ((level Int) (index Int) (level-child Int) (index-child Int)) Bool
                    (and (= level-child (+ level 1))
                         (= index-child (+ (* index 2) 1))
                         ))

(define-fun valuetree2 ((t MyList) (n Int)) Bool
                   (and
                     (exists ((o2 Obj))
                       (and
                         (select_list t 0 const_value)
                         (select_list t 1 o2)
                         (map_N o2 n)
                       )
                     ) 
                   )
)

(define-fun vartree3 ((t MyList) (e MyMap) (x Obj)) Bool
                   (and
                     (select_list t 0 const_var)
                     (select_list t 1 x)
                     (not (maps_to e x Null))
                   )
)


(define-fun optree ((t MyList) (l Obj) (r Obj) (op Obj)) Bool
                                 (and
                                   (select_list t 0 const_op)
                                   (select_list t 1 op)
                                   (select_list t 2 l)
                                   (select_list t 3 r)
                                 )
)


;def valuetree1(t:list) = exists n:num. valuetree2(t,n);
(define-fun valuetree1 ((t MyList)) Bool
                   (exists ((n Int))
                     (valuetree2 t n)
                   )
)

;def vartree2(t:list,e:map) = exists x:obj. vartree3(t,e,x);
(define-fun vartree2 ((t MyList) (e MyMap)) Bool
                   (exists ((x Obj))
                     (vartree3 t e x)
                   )
)

(define-fun optree-pt ((t MyList) (l Obj) (r Obj)) Bool
                               (exists ((op Int) (o1 Obj) (o2 Obj) (o3 Obj))
                                 (and
                                   (select_list t 0 const_op)
                                   (select_list t 1 const_add)
                                   (select_list t 2 l)
                                   (select_list t 3 r)
                                 )
                               )
)

(define-fun parsetree2 ((t MyList) (e MyMap)) Bool
                   (exists ((a (Array Int MyList)) (o (Array Int Obj)))
                     (and
                       (= (select a (index-of 0 0)) t)
                       (forall ((i Int) (li Int))
                        (and
                         (=>
                           (and (<= 0 li) (< li MAX_LEVEL) (<= 0 i) (< i (size-level li)))
                           (or
                             (valuetree1 (select a (index-of li i)))
                             (vartree2 (select a (index-of li i)) e)
                             (forall ((lj Int) (j1 Int) (j2 Int))
                               (=>
                                 (and (child-a li i lj j1) (child-b li i lj j2))
                                 (optree-pt (select a (index-of li i))
                                         (select o (index-of lj j1))
                                         (select o (index-of lj j2))
                                 )
                               )
                             )
                           )
                         )
                         (map_L (select o (index-of li i)) (select a (index-of li i)))
                        )
                       )
                     )
                   )
)


(define-fun optree5 ((t MyList) (e MyMap) (op Int) (l MyList) (r MyList)) Bool
                  (and
                    (parsetree2 t e)
                    (not (valuetree1 t))
                    (not (vartree2 t e))
                  )
)

; def optree2(t:list,e:map) = exists op:num. exists l:list. exists r:list. optree5(t,e,op,l,r);
(define-fun optree2 ((t MyList) (e MyMap)) Bool
                   (exists ((op Int) (l MyList) (r MyList))
                     (optree5 t e op l r)
                   )
)


(define-fun treeval3 ((t MyList) (e MyMap) (n Int)) Bool
    (exists ((a (Array Int MyList)) (res (Array Int Int)))
      (and
	(= (select a (index-of 0 0)) t)
	(= (select res (index-of 0 0)) n)
	(forall ((i Int) (li Int))
	  (=>
	    (and (<= 0 li) (< li MAX_LEVEL) (<= 0 i) (< i (size-level li)))
	    (exists ((lj Int) (ll Int) (rj Int) (rl Int))
	      (and
		(child-a li i ll lj)
		(child-b li i rl rj)
		(or
		  (and
		    (= (select res (index-of li i)) -1)
		    (= (select res (index-of ll lj)) -1)
		    (= (select res (index-of rl rj)) -1)
		  )
		  (and
		    (valuetree2 (select a (index-of li i)) (select res (index-of li i)))
		    (= (select res (index-of ll lj)) -1)
		    (= (select res (index-of rl rj)) -1)
		  )
		  (and
		    (exists ((x Obj) (v Obj))
		      (and
			(vartree3 (select a (index-of li i)) e x)
			(maps_to e x v)
			(map_N v (select res (index-of li i)))
		      )
		    )
		    (= (select res (index-of ll lj)) -1)
		    (= (select res (index-of rl rj)) -1)
		  )
		  (exists ((l Obj) (r Obj))
		    (and
		      (optree (select a (index-of li i)) l r const_add)
		      (map_L l (select a (index-of ll lj)))
		      (map_L r (select a (index-of rl rj)))
		      (= (select res (index-of li i))
			(+ (select res (index-of ll lj)) (select res (index-of rl rj)))
		      )
		    )
		  )
		)
	      )
	    )
	  )
	)
      )
    )
)

; =============== Testing ===============0

;#invariant ${\color{blue}\forall t:\typeALlist,e:\typeALmap \quantsep \mathit{parsetree2(t,e)} \rightarrow}$
;#       ${\color{blue}t[0] = \ProgVirtual{\PL{VALUE}} \rightarrow \mathit{valuetree1(t)} \wedge t[0] =
;# \ProgVirtual{\PL{VAR}} \rightarrow \mathit{vartree2(t,e)} \wedge t[0] = \ProgVirtual{\PL{OP}} \rightarrow
;# \mathit{optree2(t,e)}}$;

(push)

(echo "testing treeval3...")

(declare-const tree MyList)
(declare-const env MyMap)
(declare-const o Obj)

(assert (select_list tree 0 const_value))
(assert (select_list tree 1 o))
(assert (map_N o 2))

(assert (treeval3 tree env 2))

(check-sat)


