(set-option :produce-unsat-cores true)
(set-option :produce-models true)
;(set-option :produce-proofs true)

; --------------- Basic Definitions -------------------

(declare-sort Obj 0)

(declare-const Null Obj)

; --------------- Classes -------------------

(declare-sort Class 0)

(declare-fun class(Obj) Class)
(declare-const class_object Class)
(declare-const class_nullclass Class)
(declare-const class_boolean Class)
(declare-const class_numeric Class)
(declare-const class_list Class)
(declare-const class_main Class)
(assert (distinct class_object class_nullclass class_boolean class_numeric class_list class_main))
(assert (= (class Null) class_nullclass))

; --------------- Instance variables -------------------

(declare-sort IVar 0)

(declare-const iv_as_ref IVar)
(declare-const iv_pred IVar)
(declare-const iv_next IVar)
(declare-const iv_value IVar)
(declare-const iv_size IVar)
(assert (distinct iv_as_ref iv_pred iv_next iv_value iv_size))

(declare-fun inst_ref(Obj IVar) Obj)
(declare-const this Obj)

; --------------- Predicates -------------------------------


; from lib/stdlib.theory.smt:
; bool

(define-fun is_B((o Obj)) Bool (= (class o) class_boolean))

(define-fun map_B((o Obj) (b Bool)) Bool (and
                                         (is_B o)
                                         (= b (not (= (inst_ref o iv_as_ref) Null)))))


; numeric

(define-fun is_N((o Obj)) Bool (= (class o) class_numeric))

; recursive definition
;(define-fun map_N((o Obj) (n Int)) Bool (and
;                                   (is_N o)
;                                   (or
;                                     (and (= (inst_ref o iv_pred) Null) (= n 0))
;                                     (map_N (inst_ref o iv_pred) (- n 1)))))

; quantified definition
(define-fun map_N((o Obj) (n Int)) Bool
				(exists ((a_o (Array Int Obj)))
				  (and
			            (forall ((i Int))
				      (=> (and (<= 0 i) (< i n))
				        (and
				          (is_N (select a_o i))
				          (= (inst_ref (select a_o i) iv_pred) (select a_o (+ i 1)))
				        )
				      )
				    )
				    (= (select a_o 0) o)
				    (= (inst_ref (select a_o n) iv_pred) Null)
				  )))

; build-in type List

(declare-datatypes () ((MyList (mk-pair (size Int) (contents (Array Int Obj))))))

(define-fun select_list ((l MyList) (index Int) (result Obj)) Bool
                        (= (select (contents l) index) result))

(define-fun size_list ((l MyList) (size Int)) Bool
                      (= (size l) size))

(define-fun in_list ((o Obj) (l MyList)) Bool
                    (exists ((i Int)) (select_list l i o)))

; list

(define-fun is_L ((l Obj)) Bool (= (class l) class_list))

; recursive definition
;(define-fun map_L ((o Obj) (l MyList)) Bool
;                  (map_L_internal o l 0))
;
;(define-fun map_L_internal ((o Obj) (l MyList) (index Int)) Bool
;                  (or
;                    (and
;                      (= (inst_ref o iv_next) Null)
;                      (= (inst_ref o iv_value) (select l index))
;                      (size_list l (+ index 1))
;                    )
;                    (and
;                      (not (= (inst_ref o iv_next) Null))
;                      (= (inst_ref o iv_value) (select l index))
;                      (map_L_internal (inst_ref o iv_next) l (+ index 1))
;                    )
;                  )
;)

(define-fun map_L ((o Obj) (l MyList)) Bool
                           (exists ((a (Array Int Obj)))
                             (and
                               (= (select a 0) o)
                               (= (inst_ref (select a (- (size l) 1)) iv_next) Null)
                               (forall ((i Int))
                                 (=>
                                   (and (< 0 i) (< i (size l)))
                                   (= (select a i) (inst_ref (select a (- i 1)) iv_next))
                                 )
                               )
                               (forall ((i Int))
                                 (=>
                                   (and (<= 0 i) (< i (size l)))
                                   (= (select (contents l) i) (inst_ref (select a i) iv_value))
                                 )
                               )
                             )
                           )
)

(define-fun select_L ((l Obj) (i Int) (o Obj)) Bool
                     (exists ((a (Array Int Obj)))
                             (and (= (select a 0) l)
                                  (forall ((i2 Int))
                                          (=> (and (<= 0 i2) (< i2 i))
                                              (= (inst_ref (select a i2) iv_next) (select a (+ i2 1)))))
                                  (= (inst_ref (select a i) iv_value) o))))

(define-fun size_L ((l Obj) (n Int)) Bool
                   (map_N (inst_ref l iv_size) n))

; predicate in_L(o:O, l:O*) = exists i:N. select_L(i,l,o)
(define-fun in_L ((o Obj) (l Obj)) Bool
                 (exists ((i Int)) (select_L l i o)))



; from ../sorting.theory.smt:

; predicate permutation(l1:O*, l2:O*) = exists n:N. length_L(l1,n) && length_L(l2,n) && forall o:O. in_L(o,l1)
; <-> in_L(o,l2)
(define-fun permutation ((l1 MyList) (l2 MyList)) Bool
                        (forall ((o Obj)) (= (in_list o l1) (in_list o l2))))

; predicate sorted_partial(l:O*,start:N,end:N) = exists n:N. length_L(l,n) && 0 <= start < end < n &&
; forall i:N. forall j:N. start <= i < j < end implies exists ni:N. exists nj:N. map_N(l[i],ni) &&
; map_N(l[j],nj) && ni < nj
(define-fun sorted_partial ((l MyList) (start Int) (end Int)) Bool
                           (forall ((i Int) (j Int))
                                   (=> (and (<= start i) (< i j) (< j end))
                                       (exists ((ni Int) (nj Int) (oi Obj) (oj Obj))
                                               (and (select_list l i oi)
                                                    (map_N oi ni)
                                                    (select_list l j oj)
                                                    (map_N oj nj)
                                                    (< ni nj))))))


; predicate sorted(l:O*) = exists n:N. length_L(l,n) && sorted_partial(l,0,n)
(define-fun sorted ((l MyList)) Bool
                   (exists ((n Int))
                     (and
                       (size_list l n) 
                       (sorted_partial l 0 n)
                     )
                   )
)

; predicate le_L(l:O*,start:N,end:N,max:N) = forall i:N. start <= i < end implies l[i] <= max
(define-fun le_L ((l MyList) (start Int) (end Int) (omax Obj)) Bool
                 (exists ((max Int))
                 (and
                   (map_N omax max)
                   (forall ((i Int))
                         (=> (and (<= start i) (< i end))
                             (exists ((n Int) (o Obj)) (and (select_list l i o)
                                                            (map_N o n)
                                                            (<= n max)
                                                       )
                             )
                         )
                   )
                 ))
)

; predicate ge_L(l:O*,start:N,end:N,min:N) = forall i:N. start <= i < end implies min <= l[i]
(define-fun ge_L ((l MyList) (start Int) (end Int) (omin Obj)) Bool
                 (exists ((min Int))
                 (and
                   (map_N omin min)
                   (forall ((i Int))
                         (=> (and (<= start i) (< i end))
                             (exists ((n Int) (o Obj)) (and (select_list l i o)
                                                            (map_N o n)
                                                            (<= min n)
                                                       )
                             )
                         )
                   )
                 ))
)

;(define-fun prime ((p Int)) Bool
;                    (forall ((n Int))
;                      (=>
;                        (and (< 1 n) (< n p))
;                        (not (= (mod p n) 0))
;                      )
;                    )
;)

;(define-fun nth_prime ((p Int) (n Int)) Bool
;                        (exists ((a (Array Int Int)))
;                          (and
;                            (forall ((i Int))
;                              (=>
;                                (and (<= 0 i) (< i n))
;                                (prime (select a i))
;                              )
;                            )
;                            (forall ((i Int))
;                              (=>
;                                (and (<= 0 i) (< i (- n 1)))
;                                (< (select a i) (select a (+ i 1)))
;                              )
;                            )
;                            (= p (select a (- n 1)))
;                          )
;                        )
;)

;(define-fun times_of_prime_factor ((n Int) (p Int) (t Int))
;                                    (or
;                                      (and (= t 0) (= (mod n p) 0))
;                                      (and (> t 0) (= (mod n p) 0) (times_of_prime_factor (div n p) p (- t 1)))
;                                    )
;)

;(define-fun times_of_prime_factor ((n Int) (p Int) (t Int))
;                                    (exists (a (Array Int Int))
;                                      (and
;                                        (= (select a 0) n)
;                                        (forall ((i Int))
;                                          (=>
;                                            (and (<= 0 i) (< i (- t 1))
;                                            (and
;                                              (= (mod (select a i) p) 0)
;                                              (= (select a (+ i 1)) (div (select a i) p))
;                                            )
;                                          )
;                                        )
;                                      )
;                                    )
;)


;(define-fun goedel_L ((l Obj) (n Int)) Bool
;                       (goedel_L_internal l 0 n)
;)
;
;(define-fun goedel_L_internal ((l Obj) (i Int) (n Int)) Bool
;                       (or
;                         (and (= n 0) (size_L l i))
;                         (and
;                           (exists ((o Obj) (vo Int) (p Int) (n2 Int))
;                             (and 
;                               (select_L l i o)
;                               (map_N o vo)
;                               (nth_prime i p)
;                               (goedel_L_internal l (+ i 1) n2)
;                               (= n (* (exp p vo) n2))
;                             )
;                           )
;                         )
;                       )
;)

;(define-fun power ((b Int) (e Int) Int)
;                    (ite (= e 0)
;                         1
;                         (* b (power b (- e 1))
;                    )
;)

;(define-fun power ((b Int) (e Int) (r Int)) Bool
;                    (exists ((a (Array Int Int)))
;                      (and
;                        (= (select a 0) 1)
;                        (forall ((i Int))
;                          (=>
;                            (and (<= 0 i) (<= i e))
;                            (= (select a i) (* b (select a (- i 1))))
;                          )
;                        )
;                        (= r (select a e))
;                      )
;                    )
;)

;(define-fun goedel_L ((l Obj) (n Int)) Bool
;                       (exists ((a (Array Int Int)) (ls Int))
;                         (and
;                           (size_L l ls)
;                           (= (select a -1) 1)
;                           (forall ((i Int))
;                             (exists ((o Obj) (v Int) (p Int) (f Int))
;                               (=>
;                                 (and (<= 0 i) (< i ls))
;                                 (and
;                                   (select_L l i o)
;                                   (map_N o v)
;                                   (nth_prime i p)
;                                   (power p v f)
;                                   (= (select a i) (* (select a (- i 1)) f))
;                                 )
;                               )
;                             )
;                           )
;                           (= n (select a (- ls 1)))
;                         )
;                       )
;)




; --------------- Contextual Information -------------------

(declare-const lv_true Bool)
(declare-const lv_false Bool)

(assert (= lv_true true))
(assert (= lv_false false))

; TODO: local program variables


; --------------- Contextual Information -------------------
(declare-const pv_self Obj)
(assert (= (class pv_self) class_main))

; --------------- Verification Condition -------------------
(declare-const pv_result Obj)
(declare-const lv_num Int)
(declare-const pv_l Obj)
(declare-const pv_i Obj)
(declare-const pv_j Obj)
(declare-const lv_l0 MyList)
(declare-const lv_l MyList)
(declare-const lv_s Int)
(declare-const lv_n Int)
(declare-const lv_item Obj)



(assert (not (=> (and (exists ( (lv_l1 MyList))
 (exists ( (lv_i Int)) (exists ( (lv_j Int)) (and (and (and (and (and
(map_L pv_l lv_l1)
(map_N pv_i lv_i))
(map_N pv_j lv_j))
(permutation lv_l1 lv_l0))
(sorted_partial lv_l1 0 lv_i)) (exists ( (lv_o Obj)) (and (and (select_list lv_l1 lv_i lv_o)
(le_L lv_l1 0 lv_i lv_o))
(ge_L lv_l1 (+ lv_i 1) lv_j lv_o)))))))
(map_B pv_result lv_true)) (and (exists ( (lv_u62 Class)) (and (= lv_u62 (class pv_l)) (or (= lv_u62 class_list)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (and (and
(map_N pv_i lv_num)
(map_L pv_l lv_l)) (forall ( (lv__a1 Obj)) (forall ( (lv__a0 Obj)) (forall ( (lv__result Obj)) (=> 
(select_list lv_l lv_num lv__result) (and (exists ( (lv_u110 Class)) (and (= lv_u110 (class pv_l)) (or (= lv_u110 class_list)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (and (and
(map_N pv_j lv_num)
(map_L pv_l lv_l)) (forall ( (lv__result Obj)) (=> (select_list lv_l lv_num lv__result) (and (exists ( (lv_u150 Class)) (and (= lv_u150 (class lv__result)) (or (= lv_u150 class_numeric)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (and (and
(map_N lv__result lv_s)
(map_N lv__result lv_n)) (forall ( (lv__result Obj)) (=> (and (or (and (map_B lv__result lv_true) (< lv_s lv_n)) (map_B lv__result lv_false)) (not (< lv_s lv_n))) (and (exists ( (lv_u202 Class)) (and (= lv_u202 (class lv__result)) (or (= lv_u202 class_boolean)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (exists ( (lv_b Bool)) (and (map_B lv__result lv_b) (forall ( (lv__result Obj)) (=> (map_B lv__result (not lv_b)) (exists ( (lv__u2030 Bool)) (and (map_B lv__result lv__u2030) (or (and lv__u2030 (and (exists ( (lv_u253 Class)) (and (= lv_u253 (class pv_l)) (or (= lv_u253 class_list)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (and (and
(map_N pv_i lv_num)
(map_L pv_l lv_l)) (forall ( (lv__result Obj)) (=> (select_list lv_l lv_num lv__result) (and (exists ( (lv_u293 Class)) (and (= lv_u293 (class pv_l)) (or (= lv_u293 class_list)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (and (and
(map_N pv_j lv_num)
(map_L pv_l lv_l)) (forall ( (lv__result Obj)) (=> (select_list lv_l lv_num lv__result) (and (exists ( (lv_u333 Class)) (and (= lv_u333 (class pv_l)) (or (= lv_u333 class_list)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (and (and (and (map_N pv_i lv_num) (map_L pv_l lv_l)) (= lv_item lv__result)) (forall ( (lv__result Obj)) (=> (select_list lv_l lv_num lv_item) (and (exists ( (lv_u377 Class)) (and (= lv_u377 (class pv_l)) (or (= lv_u377 class_list)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (exists ( (lv_item Obj)) (exists ( (lv_l MyList)) (exists ( (lv_num Int)) (and (and (and (map_N pv_j lv_num) (map_L pv_l lv_l)) (= lv_item lv__result)) (forall ( (lv__result Obj)) (=> (select_list lv_l lv_num lv_item) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (forall ( (lv__result Obj)) (and (exists ( (lv_u445 Class)) (and (= lv_u445 (class lv__result)) (or (= lv_u445 class_numeric)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (forall ( (lv__result Obj)) (and (exists ( (lv_u472 Class)) (and (= lv_u472 (class pv_j)) (or (= lv_u472 class_numeric)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (exists ( (lv_r Int)) (and (and (and (map_N pv_j lv_s) (map_N lv__result lv_n)) (= lv_r (+ lv_s lv_n))) (forall ( (lv__result Obj)) (=> (map_N lv__result lv_r) (and (exists ( (lv_u521 Class)) (and (= lv_u521 (class pv_l)) (or (= lv_u521 class_list)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (exists ( (lv_res Obj)) (and (= lv_res (inst_ref pv_l iv_size)) (forall ( (lv__result Obj)) (=> (= lv__result lv_res) (and (exists ( (lv_u562 Class)) (and (= lv_u562 (class pv_j)) (or (= lv_u562 class_numeric)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (exists ( (lv_n Int)) (exists ( (lv_s Int)) (and (and (map_N pv_j lv_s) (map_N lv__result lv_n)) (forall ( (lv__result Obj)) (=> (and (or (and (map_B lv__result lv_true) (< lv_s lv_n)) (map_B lv__result lv_false)) (not (< lv_s lv_n))) (exists ( (lv_l1 MyList)) (exists ( (lv_i Int)) (exists ( (lv_j Int)) (and (and (and (and (and (map_L pv_l lv_l1) (map_N pv_i lv_i)) (map_N pv_j lv_j)) (permutation lv_l1 lv_l0)) (sorted_partial lv_l1 0 lv_i)) (exists ( (lv_o Obj)) (and (and (select_list lv_l1 lv_i lv_o) (le_L lv_l1 0 lv_i lv_o)) (ge_L lv_l1 (+ lv_i 1) lv_j lv_o)))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))) (and (not lv__u2030) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (forall ( (lv__result Obj)) (and (exists ( (lv_u692 Class)) (and (= lv_u692 (class lv__result)) (or (= lv_u692 class_numeric)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (forall ( (lv__result Obj)) (and (exists ( (lv_u719 Class)) (and (= lv_u719 (class pv_j)) (or (= lv_u719 class_numeric)))) (exists ( (lv_i Obj)) (exists ( (lv_j Obj)) (exists ( (lv_r Int)) (and (and (and (map_N pv_j lv_s) (map_N lv__result lv_n)) (= lv_r (+ lv_s lv_n))) (forall ( (lv__result Obj)) (=> (map_N lv__result lv_r) (and (exists ( (lv_u768 Class)) (and (= lv_u768 (class pv_l)) (or (= lv_u768 class_list)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (exists ( (lv_res Obj)) (and (= lv_res (inst_ref pv_l iv_size)) (forall ( (lv__result Obj)) (=> (= lv__result lv_res) (and (exists ( (lv_u809 Class)) (and (= lv_u809 (class pv_j)) (or (= lv_u809 class_numeric)))) (exists ( (lv_j Obj)) (exists ( (lv_i Obj)) (exists ( (lv_n Int)) (exists ( (lv_s Int)) (and (and (map_N pv_j lv_s) (map_N lv__result lv_n)) (forall ( (lv__result Obj)) (=> (and (or (and (map_B lv__result lv_true) (< lv_s lv_n)) (map_B lv__result lv_false)) (not (< lv_s lv_n))) (exists ( (lv_l1 MyList)) (exists ( (lv_i Int)) (exists ( (lv_j Int)) (and (and (and (and (and (map_L pv_l lv_l1) (map_N pv_i lv_i)) (map_N pv_j lv_j)) (permutation lv_l1 lv_l0)) (sorted_partial lv_l1 0 lv_i)) (exists ( (lv_o Obj)) (and (and (select_list lv_l1 lv_i lv_o) (le_L lv_l1 0 lv_i lv_o)) (ge_L lv_l1 (+ lv_i 1) lv_j lv_o))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))))

; --------------- Verify -------------------
(check-sat)
