﻿namespace Dyn.Console.CommandLine
{
    using System;
    using System.Runtime.Serialization;

    /// <summary>
    /// The runtime exception that is thrown when <see cref="CommandLineParser{T}.Parse(string[], T)"/> is unable to interpret a command line argument.
    /// </summary>
    [Serializable]
    public sealed class ArgumentParserException : Exception
    {
        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentParserException"/> class.
        /// </summary>
        public ArgumentParserException()
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentParserException"/> class with a specified error message.
        /// </summary>
        /// <param name="message">The message that describes the error.</param>
        public ArgumentParserException(string message)
            : base(message)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentParserException"/> class with a specified error message and a reference to the inner exception that is the cause of this exception.
        /// </summary>
        /// <param name="message">The error message that explains the reason for the exception.</param>
        /// <param name="inner">The <see cref="Exception"/> that is the cause of the current <see cref="ArgumentParserException"/>, or a <code>null</code> reference (<code>Nothing</code> in Visual Basic) if no inner exception is specified.</param>
        public ArgumentParserException(string message, Exception inner)
            : base(message, inner)
        {
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="ArgumentParserException"/> class with serialized data.
        /// </summary>
        /// <param name="info">The object that holds the serialized object data.</param>
        /// <param name="context">The contextual information about the source or destination.</param>
        private ArgumentParserException(
          SerializationInfo info,
          StreamingContext context)
            : base(info, context)
        {
        }

        /// <summary>
        /// Populates a <see cref="SerializationInfo"/> with the data needed to serialize the target object.
        /// </summary>
        /// <param name="info">The <see cref="SerializationInfo"/> to populate with data.</param>
        /// <param name="context">The destination (see <see cref="StreamingContext"/>) for this serialization.</param>
        public override void GetObjectData(SerializationInfo info, StreamingContext context)
        {
            base.GetObjectData(info, context);
        }
    }
}
