﻿namespace Dyn.Console.CommandLine
{
    using System;

    /// <summary>
    /// Represents an attribute used to define a command line interface and associate it with properties, fields or methods.
    /// </summary>
    [AttributeUsage(
        AttributeTargets.Method |
        AttributeTargets.Field |
        AttributeTargets.Property,
        Inherited = false,
        AllowMultiple = true)]
    internal sealed class OptionAttribute : Attribute
    {
        /// <summary>
        /// Gets or sets the short form of the option to be accessed from command line.
        /// </summary>
        public char ShortName { get; set; }

        /// <summary>
        /// Gets or sets the long form of the option to be accessed from command line.
        /// </summary>
        public string LongName { get; set; }

        /// <summary>
        /// Gets or sets the description of the command line option.
        /// </summary>
        public string Description { get; set; }

        /// <summary>
        ///  Gets or sets a value indicating whether the associated member to handles arguments.
        /// </summary>
        /// <remarks>At most one <see cref="OptionAttribute"/> per class or structure is allowed to be the argument.</remarks>
        public bool IsArgument { get; set; }
    }
}
