﻿namespace Dyn.Console
{
    using System;
    using System.Collections.Generic;
    using Dyn.Diagnostics;

    /// <summary>
    /// Counts diagnostics events.
    /// </summary>
    /// <remarks>This class support event chaining.</remarks>
    public sealed class DiagnosticSummary : IDiagnostic
    {
        /// <summary>
        /// Counter
        /// </summary>
        private readonly Dictionary<DiagnosticSeverity, int> _counter =
            new Dictionary<DiagnosticSeverity, int>();

        /// <inheritdoc/>
        public event EventHandler<DiagnosticEventArgs> DiagnosticInfo;

        /// <summary>
        /// Returns the number of occurrences of <see cref="DiagnosticEventArgs"/> events with given <see cref="DiagnosticSeverity"/>.
        /// </summary>
        /// <param name="key">The severity</param>
        /// <returns>The number of events with that severity.</returns>
        public int this[DiagnosticSeverity key]
        {
            get
            {
                int result;

                if (!this._counter.TryGetValue(key, out result))
                {
                    result = 0;
                }

                return result;
            }
        }

        /// <summary>
        /// Counts events based on their severity and raises the same event as <see cref="IDiagnostic.DiagnosticInfo"/>.
        /// </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="args">The arguments.</param>
        public void DiagnosticInfoListener(object sender, DiagnosticEventArgs args)
        {
            DiagnosticSeverity key = args.Info.Level;

            this._counter[key] = this[key] + 1;

            var handler = this.DiagnosticInfo;
            if (handler != null)
            {
                handler(sender, args);
            }
        }
    }
}
