﻿using Dyn.Lexer;
using System;
using System.Collections.Generic;

// Generated code from Generator/DynAst
namespace Dyn.Ast
{
    //
    // Assertions
    //
    //
    public abstract class DynAssertionExpressionNode
    {
        public enum NodeKind
        {
            Addition,
            And,
            AssertionVariable,
            Division,
            Equality,
            Existential,
            GreaterThan,
            GreaterThanOrEqual,
            IfThenElse,
            Implies,
            InstanceVariable,
            LessThan,
            LessThanOrEqual,
            Modulus,
            Multiplication,
            Negate,
            Numeric,
            Or,
            Predicate,
            ProgramVariable,
            String,
            Subtraction,
            TypeCheck,
            Universal
        }

        private readonly Span _location;
        private readonly NodeKind _kind;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionExpressionNode" /> class.</summary>
        protected internal DynAssertionExpressionNode(Span location, NodeKind kind)
        {
            this._location = location;
            this._kind = kind;
        }

        public Span Location { get { return this._location; } }

        public NodeKind Kind { get { return this._kind; } }
    }

    public sealed class DynAssertionQuantifierNode : DynAssertionExpressionNode
    {
        private readonly string _variable;
        private readonly string _type;
        private readonly DynAssertionExpressionNode _expression;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionQuantifierNode" /> class.</summary>
        public DynAssertionQuantifierNode(Span location, NodeKind kind, string variable, string type, DynAssertionExpressionNode expression)
                : base(location, ValidateKind(kind))
        {
            Guard.EnsureNotNull("variable", variable);
            Guard.EnsureNotNull("type", type);
            Guard.EnsureNotNull("expression", expression);

            this._variable = variable;
            this._type = type;
            this._expression = expression;
        }

        public static NodeKind ValidateKind(NodeKind kind)
        {
            switch (kind)
            {
                case NodeKind.Universal:
                case NodeKind.Existential:
                    return kind;

                default:
                    throw Guard.EnumOutOfRangeException<NodeKind>("kind", (int)kind);
            }
        }

        public string Variable { get { return this._variable; } }

        public string Type { get { return this._type; } }

        public DynAssertionExpressionNode Expression { get { return this._expression; } }
    }

    public sealed class DynAssertionTypeCheckNode : DynAssertionExpressionNode
    {
        private readonly DynAssertionExpressionNode _expression;
        private readonly ICollection<string> _typeSet;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionTypeCheckNode" /> class.</summary>
        public DynAssertionTypeCheckNode(Span location, DynAssertionExpressionNode expression, ICollection<string> typeSet)
                : base(location, NodeKind.TypeCheck)
        {
            Guard.EnsureNotNull("expression", expression);

            this._expression = expression;
            this._typeSet = Utilities.AsImmutableList(typeSet);
        }

        public DynAssertionExpressionNode Expression { get { return this._expression; } }

        public ICollection<string> TypeSet { get { return this._typeSet; } }

    }

    public sealed class DynAssertionPredicateNode : DynAssertionExpressionNode
    {
        private readonly string _name;
        private readonly IList<DynAssertionExpressionNode> _arguments;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionPredicateNode" /> class.</summary>
        public DynAssertionPredicateNode(Span location, string name, IList<DynAssertionExpressionNode> arguments)
                : base(location, NodeKind.Predicate)
        {
            Guard.EnsureNotNull("name", name);

            this._name = name;
            this._arguments = Utilities.AsImmutableList(arguments);
        }

        public string Name { get { return this._name; } }

        public IList<DynAssertionExpressionNode> Arguments { get { return this._arguments; } }
    }

    public sealed class DynAssertionUnaryNode : DynAssertionExpressionNode
    {
        private readonly DynAssertionExpressionNode _expression;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionUnaryNode" /> class.</summary>
        public DynAssertionUnaryNode(Span location, DynAssertionExpressionNode expression)
                : base(location, NodeKind.Negate)
        {
            Guard.EnsureNotNull("expression", expression);

            this._expression = expression;
        }

        public DynAssertionExpressionNode Expression { get { return this._expression; } }
    }

    public sealed class DynAssertionBinaryNode : DynAssertionExpressionNode
    {
        private readonly DynAssertionExpressionNode _first;
        private readonly DynAssertionExpressionNode _second;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionBinaryNode" /> class.</summary>
        public DynAssertionBinaryNode(Span location, NodeKind kind, DynAssertionExpressionNode first, DynAssertionExpressionNode second)
                : base(location, ValidateKind(kind))
        {
            Guard.EnsureNotNull("first", first);
            Guard.EnsureNotNull("second", second);

            this._first = first;
            this._second = second;
        }

        public static NodeKind ValidateKind(NodeKind kind)
        {
            switch (kind)
            {
                case NodeKind.And:
                case NodeKind.Or:
                case NodeKind.Implies:
                case NodeKind.Addition:
                case NodeKind.Subtraction:
                case NodeKind.Multiplication:
                case NodeKind.Division:
                case NodeKind.Modulus:
                case NodeKind.GreaterThan:
                case NodeKind.GreaterThanOrEqual:
                case NodeKind.LessThan:
                case NodeKind.LessThanOrEqual:
                case NodeKind.Equality:
                    return kind;

                default:
                    throw Guard.EnumOutOfRangeException<NodeKind>("kind", (int)kind);
            }
        }

        public DynAssertionExpressionNode First { get { return this._first; } }

        public DynAssertionExpressionNode Second { get { return this._second; } }
    }

    public sealed class DynAssertionIfThenElseNode : DynAssertionExpressionNode
    {
        private readonly DynAssertionExpressionNode _condition;
        private readonly DynAssertionExpressionNode _trueBranch;
        private readonly DynAssertionExpressionNode _falseBranch;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionIfThenElseNode" /> class.</summary>
        public DynAssertionIfThenElseNode(Span location, DynAssertionExpressionNode condition, DynAssertionExpressionNode trueBranch, DynAssertionExpressionNode falseBranch)
                : base(location, NodeKind.IfThenElse)
        {
            Guard.EnsureNotNull("condition", condition);
            Guard.EnsureNotNull("trueBranch", trueBranch);
            Guard.EnsureNotNull("falseBranch", falseBranch);

            this._condition = condition;
            this._trueBranch = trueBranch;
            this._falseBranch = falseBranch;
        }

        public DynAssertionExpressionNode Condition { get { return this._condition; } }

        public DynAssertionExpressionNode TrueBranch { get { return this._trueBranch; } }

        public DynAssertionExpressionNode FalseBranch { get { return this._falseBranch; } }
    }

    public sealed class DynAssertionNumericNode : DynAssertionExpressionNode
    {
        private readonly int _value;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionNumericNode" /> class.</summary>
        public DynAssertionNumericNode(Span location, int value)
                : base(location, NodeKind.Numeric)
        {
            this._value = value;
        }

        public int Value { get { return this._value; } }
    }

    public sealed class DynAssertionStringNode : DynAssertionExpressionNode
    {
        private readonly string _value;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionStringNode" /> class.</summary>
        public DynAssertionStringNode(Span location, string value)
                : base(location, NodeKind.String)
        {
            Guard.EnsureNotNull("value", value);

            this._value = value;
        }

        public string Value { get { return this._value; } }
    }

    public sealed class DynAssertionInstanceVariableNode : DynAssertionExpressionNode
    {
        private readonly DynAssertionExpressionNode _expression;
        private readonly string _name;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionInstanceVariableNode" /> class.</summary>
        public DynAssertionInstanceVariableNode(Span location, DynAssertionExpressionNode expression, string name)
                : base(location, NodeKind.InstanceVariable)
        {
            Guard.EnsureNotNull("expression", expression);
            Guard.EnsureNotNull("name", name);

            this._expression = expression;
            this._name = name;
        }

        public DynAssertionExpressionNode Expression { get { return this._expression; } }

        public string Name { get { return this._name; } }
    }

    public sealed class DynAssertionVariableNode : DynAssertionExpressionNode
    {
        private readonly string _name;

        /// <summary> Initializes a new instance of the <see cref="DynAssertionVariableNode" /> class.</summary>
        public DynAssertionVariableNode(Span location, NodeKind kind, string name)
                : base(location, ValidateKind(kind))
        {
            Guard.EnsureNotNull("name", name);

            this._name = name;
        }

        public static NodeKind ValidateKind(NodeKind kind)
        {
            switch (kind)
            {
                case NodeKind.AssertionVariable:
                case NodeKind.ProgramVariable:
                    return kind;

                default:
                    throw Guard.EnumOutOfRangeException<NodeKind>("kind", (int)kind);
            }
        }

        public string Name { get { return this._name; } }
    }

    public class DynAssertionNodeTransformer
    {
        public virtual DynAssertionExpressionNode Transform(DynAssertionExpressionNode node)
        {
            Guard.EnsureNotNull("node", node);

            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return Transform((DynAssertionQuantifierNode)node);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return Transform((DynAssertionTypeCheckNode)node);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return Transform((DynAssertionPredicateNode)node);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return Transform((DynAssertionUnaryNode)node);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    return Transform((DynAssertionBinaryNode)node);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return Transform((DynAssertionIfThenElseNode)node);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return Transform((DynAssertionNumericNode)node);

                case DynAssertionExpressionNode.NodeKind.String:
                    return Transform((DynAssertionStringNode)node);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return Transform((DynAssertionInstanceVariableNode)node);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return Transform((DynAssertionVariableNode)node);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionQuantifierNode node)
        {
            return new DynAssertionQuantifierNode(
                    node.Location,
                    node.Kind,
                    node.Variable,
                    node.Type,
                    Transform(node.Expression));
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionTypeCheckNode node)
        {
            return new DynAssertionTypeCheckNode(
                    node.Location,
                    Transform(node.Expression),
                    node.TypeSet);
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionPredicateNode node)
        {
            return new DynAssertionPredicateNode(
                    node.Location,
                    node.Name,
                    Utilities.Transform(node.Arguments, Transform));
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionUnaryNode node)
        {
            return new DynAssertionUnaryNode(
                    node.Location,
                    Transform(node.Expression));
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionBinaryNode node)
        {
            return new DynAssertionBinaryNode(
                    node.Location,
                    node.Kind,
                    Transform(node.First),
                    Transform(node.Second));
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionIfThenElseNode node)
        {
            return new DynAssertionIfThenElseNode(
                    node.Location,
                    Transform(node.Condition),
                    Transform(node.TrueBranch),
                    Transform(node.FalseBranch));
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionNumericNode node)
        {
            return new DynAssertionNumericNode(
                    node.Location,
                    node.Value);
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionStringNode node)
        {
            return new DynAssertionStringNode(
                    node.Location,
                    node.Value);
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionInstanceVariableNode node)
        {
            return new DynAssertionInstanceVariableNode(
                    node.Location,
                    Transform(node.Expression),
                    node.Name);
        }

        public virtual DynAssertionExpressionNode Transform(DynAssertionVariableNode node)
        {
            return new DynAssertionVariableNode(
                    node.Location,
                    node.Kind,
                    node.Name);
        }
    }

    //
    // Program
    //
    //
    public sealed class DynProgramClassNode
    {
        private readonly Span _location;
        private readonly string _name;
        private readonly string _parent;
        private readonly IList<DynProgramRenameDeclarationNode> _renames;
        private readonly IList<DynProgramMethodDeclarationNode> _methods;
        private readonly IList<DynProgramContractDeclarationNode> _contracts;

        /// <summary> Initializes a new instance of the <see cref="DynProgramClassNode" /> class.</summary>
        public DynProgramClassNode(Span location, string name, string parent, IList<DynProgramRenameDeclarationNode> renames, IList<DynProgramMethodDeclarationNode> methods, IList<DynProgramContractDeclarationNode> contracts)
        {
            Guard.EnsureNotNull("name", name);
            Guard.EnsureNotNull("parent", parent);

            this._location = location;
            this._name = name;
            this._parent = parent;
            this._renames = Utilities.AsImmutableList(renames);
            this._methods = Utilities.AsImmutableList(methods);
            this._contracts = Utilities.AsImmutableList(contracts);
        }

        public Span Location { get { return this._location; } }

        public string Name { get { return this._name; } }

        public string Parent { get { return this._parent; } }

        public IList<DynProgramRenameDeclarationNode> Renames { get { return this._renames; } }

        public IList<DynProgramMethodDeclarationNode> Methods { get { return this._methods; } }

        public IList<DynProgramContractDeclarationNode> Contracts { get { return this._contracts; } }
    }

    public sealed class DynProgramRenameDeclarationNode
    {
        private readonly Span _location;
        private readonly string _from;
        private readonly string _to;

        /// <summary> Initializes a new instance of the <see cref="DynProgramRenameDeclarationNode" /> class.</summary>
        public DynProgramRenameDeclarationNode(Span location, string from, string to)
        {
            Guard.EnsureNotNull("from", from);
            Guard.EnsureNotNull("to", to);

            this._location = location;
            this._from = from;
            this._to = to;
        }

        public Span Location { get { return this._location; } }

        public string From { get { return this._from; } }

        public string To { get { return this._to; } }
    }

    public sealed class DynProgramMethodDeclarationNode
    {
        private readonly Span _location;
        private readonly string _name;
        private readonly IList<string> _parameters;
        private readonly IList<DynProgramStatementNode> _statements;
        private readonly IList<DynProgramContractDeclarationNode> _contracts;

        /// <summary> Initializes a new instance of the <see cref="DynProgramMethodDeclarationNode" /> class.</summary>
        public DynProgramMethodDeclarationNode(Span location, string name, IList<string> parameters, IList<DynProgramStatementNode> statements, IList<DynProgramContractDeclarationNode> contracts)
        {
            Guard.EnsureNotNull("name", name);
            Guard.EnsureNotNull("contracts", contracts);

            this._location = location;
            this._name = name;
            this._parameters = Utilities.AsImmutableList(parameters);
            this._statements = Utilities.AsImmutableList(statements);
            this._contracts = Utilities.AsImmutableList(contracts);
        }

        public Span Location { get { return this._location; } }

        public string Name { get { return this._name; } }

        public IList<string> Parameters { get { return this._parameters; } }

        public IList<DynProgramStatementNode> Statements { get { return this._statements; } }

        public IList<DynProgramContractDeclarationNode> Contracts { get { return this._contracts; } }
    }

    public abstract class DynProgramStatementNode
    {
        public enum NodeKind
        {
            Assert,
            Assignment,
            ByTheWay,
            ConstructorCall,
            ObjectCreation,
            Ensures,
            If,
            InstanceVariable,
            Invariant,
            MethodCall,
            Requires,
            Variable,
            While
        }

        private readonly Span _location;
        private readonly NodeKind _kind;

        /// <summary> Initializes a new instance of the <see cref="DynProgramStatementNode" /> class.</summary>
        protected internal DynProgramStatementNode(Span location, NodeKind kind)
        {
            Guard.EnsureEnumDefined<NodeKind>("kind", (int)kind);

            this._location = location;
            this._kind = kind;
        }

        public Span Location { get { return this._location; } }

        public NodeKind Kind { get { return this._kind; } }
    }

    public sealed class DynProgramContractDeclarationNode : DynProgramStatementNode
    {
        private readonly DynAssertionExpressionNode _expression;

        /// <summary> Initializes a new instance of the <see cref="DynProgramContractDeclarationNode" /> class.</summary>
        public DynProgramContractDeclarationNode(Span location, NodeKind kind, DynAssertionExpressionNode expression)
                : base(location, ValidateKind(kind))
        {
            Guard.EnsureNotNull("expression", expression);

            this._expression = expression;
        }

        public static NodeKind ValidateKind(NodeKind kind)
        {
            switch (kind)
            {
                case NodeKind.Ensures:
                case NodeKind.Requires:
                case NodeKind.Invariant:
                case NodeKind.Assert:
                case NodeKind.ByTheWay:
                    return kind;

                default:
                    throw Guard.EnumOutOfRangeException<NodeKind>("kind", (int)kind);
            }
        }

        public DynAssertionExpressionNode Expression { get { return this._expression; } }
    }

    public abstract class DynProgramExpressionNode : DynProgramStatementNode
    {
        /// <summary> Initializes a new instance of the <see cref="DynProgramExpressionNode" /> class.</summary>
        protected internal DynProgramExpressionNode(Span location, NodeKind kind)
                : base(location, kind)
        {
        }
    }

    public sealed class DynProgramMethodCallNode : DynProgramExpressionNode
    {
        private readonly DynProgramExpressionNode _expression;
        private readonly string _name;
        private readonly IList<DynProgramExpressionNode> _arguments;

        /// <summary> Initializes a new instance of the <see cref="DynProgramMethodCallNode" /> class.</summary>
        public DynProgramMethodCallNode(Span location, DynProgramExpressionNode expression, string name, IList<DynProgramExpressionNode> arguments)
                : base(location, NodeKind.MethodCall)
        {
            Guard.EnsureNotNull("expression", expression);
            Guard.EnsureNotNull("name", name);

            this._expression = expression;
            this._name = name;
            this._arguments = Utilities.AsImmutableList(arguments);
        }

        public DynProgramExpressionNode Expression { get { return this._expression; } }

        public string Name { get { return this._name; } }

        public IList<DynProgramExpressionNode> Arguments { get { return this._arguments; } }
    }

    public sealed class DynProgramConstructorCallNode : DynProgramExpressionNode
    {
        private readonly string _name;
        private readonly IList<DynProgramExpressionNode> _arguments;

        /// <summary> Initializes a new instance of the <see cref="DynProgramConstructorCallNode" /> class.</summary>
        public DynProgramConstructorCallNode(Span location, string name, IList<DynProgramExpressionNode> arguments)
                : base(location, NodeKind.ConstructorCall)
        {
            Guard.EnsureNotNull("name", name);

            this._name = name;
            this._arguments = Utilities.AsImmutableList(arguments);
        }

        public string Name { get { return this._name; } }

        public IList<DynProgramExpressionNode> Arguments { get { return this._arguments; } }
    }

    public sealed class DynProgramObjectCreationNode : DynProgramExpressionNode
    {
        private readonly string _classname;

        /// <summary> Initializes a new instance of the <see cref="DynProgramObjectCreationNode" /> class.</summary>
        public DynProgramObjectCreationNode(Span location, string classname)
                : base(location, NodeKind.ObjectCreation)
        {
            Guard.EnsureNotNull("classname", classname);

            this._classname = classname;
        }

        public string ClassName { get { return this._classname; } }
    }

    public sealed class DynProgramVariableNode : DynProgramExpressionNode
    {
        private readonly string _name;

        /// <summary> Initializes a new instance of the <see cref="DynProgramVariableNode" /> class.</summary>
        public DynProgramVariableNode(Span location, NodeKind kind, string name)
                : base(location, ValidateKind(kind))
        {
            Guard.EnsureNotNull("name", name);

            this._name = name;
        }

        public static NodeKind ValidateKind(NodeKind kind)
        {
            switch (kind)
            {
                case NodeKind.Variable:
                case NodeKind.InstanceVariable:
                    return kind;

                default:
                    throw Guard.EnumOutOfRangeException<NodeKind>("kind", (int)kind);
            }
        }

        public string Name { get { return this._name; } }
    }

    public sealed class DynProgramIfNode : DynProgramExpressionNode
    {
        private readonly DynProgramExpressionNode _condition;
        private readonly IList<DynProgramStatementNode> _trueBranch;
        private readonly IList<DynProgramStatementNode> _falseBranch;

        /// <summary> Initializes a new instance of the <see cref="DynProgramIfNode" /> class.</summary>
        public DynProgramIfNode(Span location, DynProgramExpressionNode condition, IList<DynProgramStatementNode> trueBranch, IList<DynProgramStatementNode> falseBranch)
                : base(location, NodeKind.If)
        {
            Guard.EnsureNotNull("condition", condition);

            this._condition = condition;
            this._trueBranch = Utilities.AsImmutableList(trueBranch);
            this._falseBranch = Utilities.AsImmutableList(falseBranch);
        }

        public DynProgramExpressionNode Condition { get { return this._condition; } }

        public IList<DynProgramStatementNode> TrueBranch { get { return this._trueBranch; } }

        public IList<DynProgramStatementNode> FalseBranch { get { return this._falseBranch; } }
    }

    public sealed class DynProgramWhileNode : DynProgramExpressionNode
    {
        private readonly DynProgramExpressionNode _condition;
        private readonly DynAssertionExpressionNode _invariant;
        private readonly IList<DynProgramStatementNode> _body;

        /// <summary> Initializes a new instance of the <see cref="DynProgramWhileNode" /> class.</summary>
        public DynProgramWhileNode(Span location, DynProgramExpressionNode condition, DynAssertionExpressionNode invariant, IList<DynProgramStatementNode> body)
                : base(location, NodeKind.While)
        {
            Guard.EnsureNotNull("condition", condition);
            Guard.EnsureNotNull("invariant", invariant);

            this._condition = condition;
            this._invariant = invariant;
            this._body = Utilities.AsImmutableList(body);
        }

        public DynProgramExpressionNode Condition { get { return this._condition; } }

        public DynAssertionExpressionNode Invariant { get { return this._invariant; } }

        public IList<DynProgramStatementNode> Body { get { return this._body; } }
    }

    public sealed class DynProgramAssignmentNode : DynProgramExpressionNode
    {
        private readonly DynProgramVariableNode _variable;
        private readonly DynProgramExpressionNode _expression;

        /// <summary> Initializes a new instance of the <see cref="DynProgramAssignmentNode" /> class.</summary>
        public DynProgramAssignmentNode(Span location, DynProgramVariableNode variable, DynProgramExpressionNode expression)
                : base(location, NodeKind.Assignment)
        {
            Guard.EnsureNotNull("variable", variable);
            Guard.EnsureNotNull("expression", expression);

            this._variable = variable;
            this._expression = expression;
        }

        public DynProgramVariableNode Variable { get { return this._variable; } }

        public DynProgramExpressionNode Expression { get { return this._expression; } }
    }

    public class DynProgramNodeTransformer
    {
        public virtual DynProgramClassNode Transform(DynProgramClassNode node)
        {
            Guard.EnsureNotNull("node", node);

            return new DynProgramClassNode(
                    node.Location,
                    node.Name,
                    node.Parent,
                    Utilities.Transform(node.Renames, Transform),
                    Utilities.Transform(node.Methods, Transform),
                    Utilities.Transform(node.Contracts, Transform));
        }

        public virtual DynProgramRenameDeclarationNode Transform(DynProgramRenameDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);

            return new DynProgramRenameDeclarationNode(
                    node.Location,
                    node.From,
                    node.To);
        }

        public virtual DynProgramMethodDeclarationNode Transform(DynProgramMethodDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);

            return new DynProgramMethodDeclarationNode(
                    node.Location,
                    node.Name,
                    node.Parameters,
                    Utilities.Transform(node.Statements, Transform),
                    Utilities.Transform(node.Contracts, Transform));
        }

        public virtual DynProgramStatementNode Transform(DynProgramStatementNode node)
        {
            Guard.EnsureNotNull("node", node);

            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return Transform((DynProgramContractDeclarationNode)node);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Transform((DynProgramMethodCallNode)node);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Transform((DynProgramConstructorCallNode)node);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Transform((DynProgramVariableNode)node);

                case DynProgramStatementNode.NodeKind.If:
                    return Transform((DynProgramIfNode)node);

                case DynProgramStatementNode.NodeKind.While:
                    return Transform((DynProgramWhileNode)node);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Transform((DynProgramAssignmentNode)node);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        public virtual DynProgramContractDeclarationNode Transform(DynProgramContractDeclarationNode node)
        {
            Guard.EnsureNotNull("node", node);

            return new DynProgramContractDeclarationNode(
                    node.Location,
                    node.Kind,
                    node.Expression);
        }

        public virtual DynProgramExpressionNode Transform(DynProgramExpressionNode node)
        {
            Guard.EnsureNotNull("node", node);

            switch (node.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Transform((DynProgramMethodCallNode)node);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Transform((DynProgramConstructorCallNode)node);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Transform((DynProgramVariableNode)node);

                case DynProgramStatementNode.NodeKind.If:
                    return Transform((DynProgramIfNode)node);

                case DynProgramStatementNode.NodeKind.While:
                    return Transform((DynProgramWhileNode)node);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Transform((DynProgramAssignmentNode)node);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("node.Kind", (int)node.Kind);
            }
        }

        public virtual DynProgramExpressionNode Transform(DynProgramMethodCallNode node)
        {
            return new DynProgramMethodCallNode(
                    node.Location,
                    Transform(node.Expression),
                    node.Name,
                    Utilities.Transform(node.Arguments, Transform));
        }

        public virtual DynProgramExpressionNode Transform(DynProgramConstructorCallNode node)
        {
            return new DynProgramConstructorCallNode(
                    node.Location,
                    node.Name,
                    Utilities.Transform(node.Arguments, Transform));
        }

        public virtual DynProgramVariableNode Transform(DynProgramVariableNode node)
        {
            Guard.EnsureNotNull("node", node);

            return new DynProgramVariableNode(
                    node.Location,
                    node.Kind,
                    node.Name);
        }

        public virtual DynProgramExpressionNode Transform(DynProgramIfNode node)
        {
            return new DynProgramIfNode(
                    node.Location,
                    Transform(node.Condition),
                    Utilities.Transform(node.TrueBranch, Transform),
                    Utilities.Transform(node.FalseBranch, Transform));
        }

        public virtual DynProgramExpressionNode Transform(DynProgramWhileNode node)
        {
            return new DynProgramWhileNode(
                    node.Location,
                    Transform(node.Condition),
                    node.Invariant,
                    Utilities.Transform(node.Body, Transform));
        }

        public virtual DynProgramExpressionNode Transform(DynProgramAssignmentNode node)
        {
            return new DynProgramAssignmentNode(
                    node.Location,
                    Transform(node.Variable),
                    Transform(node.Expression));
        }
    }

    //
    // Hoare
    //
    //
    /// <summary>A hoare triple containing a precondition, zero or more statements and a postcondition.</summary>
    /// <remarks>This class is immutable and therefore thread-safe.</remarks>
    public sealed class DynHoareNode
    {
        private readonly Span _location;
        private readonly DynAssertionExpressionNode _preCondition;
        private readonly IList<DynProgramStatementNode> _statements;
        private readonly DynAssertionExpressionNode _postCondition;

        /// <summary> Initializes a new instance of the <see cref="DynHoareNode" /> class.</summary>
        /// <param name="Location">The source code position of this node.</param>
        /// <param name="PreCondition">The precondition.</param>
        /// <param name="Statements">The statements.</param>
        /// <param name="PostCondition">The postcondition.</param>
        public DynHoareNode(Span location, DynAssertionExpressionNode preCondition, IList<DynProgramStatementNode> statements, DynAssertionExpressionNode postCondition)
        {
            this._location = location;
            this._preCondition = preCondition ?? AstExtensions.Assertions.True;
            this._statements = Utilities.AsImmutableList(statements);
            this._postCondition = postCondition ?? AstExtensions.Assertions.True;
        }

        /// <summary> Gets or sets the source code position of this node.</summary>
        /// <value>The source code position of this node.</value>
        /// <returns>The source code position of this node.</returns>
        public Span Location { get { return this._location; } }

        /// <summary> Gets or sets the precondition.</summary>
        /// <value>The precondition.</value>
        /// <returns>The precondition.</returns>
        public DynAssertionExpressionNode PreCondition { get { return this._preCondition; } }

        /// <summary> Gets or sets the statements.</summary>
        /// <value>The statements.</value>
        /// <returns>The statements.</returns>
        public IList<DynProgramStatementNode> Statements { get { return this._statements; } }

        /// <summary> Gets or sets the postcondition.</summary>
        /// <value>The postcondition.</value>
        /// <returns>The postcondition.</returns>
        public DynAssertionExpressionNode PostCondition { get { return this._postCondition; } }
    }

    //
    // EqualityComparer with Location
    //
    //
    public sealed class DynNodeEqualityComparer :
                    IEqualityComparer<DynAssertionExpressionNode>,
                    IEqualityComparer<DynAssertionQuantifierNode>,
                    IEqualityComparer<DynAssertionTypeCheckNode>,
                    IEqualityComparer<DynAssertionPredicateNode>,
                    IEqualityComparer<DynAssertionUnaryNode>,
                    IEqualityComparer<DynAssertionBinaryNode>,
                    IEqualityComparer<DynAssertionIfThenElseNode>,
                    IEqualityComparer<DynAssertionNumericNode>,
                    IEqualityComparer<DynAssertionStringNode>,
                    IEqualityComparer<DynAssertionInstanceVariableNode>,
                    IEqualityComparer<DynAssertionVariableNode>,
                    IEqualityComparer<DynProgramClassNode>,
                    IEqualityComparer<DynProgramRenameDeclarationNode>,
                    IEqualityComparer<DynProgramMethodDeclarationNode>,
                    IEqualityComparer<DynProgramStatementNode>,
                    IEqualityComparer<DynProgramContractDeclarationNode>,
                    IEqualityComparer<DynProgramExpressionNode>,
                    IEqualityComparer<DynProgramMethodCallNode>,
                    IEqualityComparer<DynProgramConstructorCallNode>,
                    IEqualityComparer<DynProgramVariableNode>,
                    IEqualityComparer<DynProgramIfNode>,
                    IEqualityComparer<DynProgramWhileNode>,
                    IEqualityComparer<DynProgramAssignmentNode>,
                    IEqualityComparer<DynHoareNode>
    {
        private static readonly IEqualityComparer<string> STRING_COMPARER = StringComparer.Ordinal;
        private const int HASH_PRIME = Utilities.HASH_PRIME;
        private const int HASH_SEED = Utilities.HASH_SEED;

        private static int GetHashCode(string value)
        {
            return value != null ? value.GetHashCode() : 0;
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionExpressionNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionExpressionNode obj)
        {
            switch (obj.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return GetHashCode((DynAssertionQuantifierNode)obj);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return GetHashCode((DynAssertionTypeCheckNode)obj);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return GetHashCode((DynAssertionPredicateNode)obj);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return GetHashCode((DynAssertionUnaryNode)obj);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    return GetHashCode((DynAssertionBinaryNode)obj);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return GetHashCode((DynAssertionIfThenElseNode)obj);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return GetHashCode((DynAssertionNumericNode)obj);

                case DynAssertionExpressionNode.NodeKind.String:
                    return GetHashCode((DynAssertionStringNode)obj);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynAssertionInstanceVariableNode)obj);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return GetHashCode((DynAssertionVariableNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionQuantifierNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionQuantifierNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                result = result * HASH_PRIME + GetHashCode(obj.Type);
                result = result * HASH_PRIME + GetHashCode(obj.Variable);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionTypeCheckNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionTypeCheckNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.TypeSet, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionPredicateNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionPredicateNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionUnaryNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionUnaryNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionBinaryNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionBinaryNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Second);
                result = result * HASH_PRIME + GetHashCode(obj.First);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionIfThenElseNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionIfThenElseNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.FalseBranch);
                result = result * HASH_PRIME + GetHashCode(obj.TrueBranch);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionNumericNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionNumericNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + obj.Value.GetHashCode();
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionStringNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionStringNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Value);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionInstanceVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionInstanceVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramClassNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramClassNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Contracts, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Methods, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Renames, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Parent);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramRenameDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramRenameDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + GetHashCode(obj.To);
                result = result * HASH_PRIME + GetHashCode(obj.From);
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramMethodDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramMethodDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Contracts, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Statements, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Parameters, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramStatementNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramStatementNode obj)
        {
            switch (obj.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return GetHashCode((DynProgramContractDeclarationNode)obj);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return GetHashCode((DynProgramMethodCallNode)obj);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return GetHashCode((DynProgramConstructorCallNode)obj);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynProgramVariableNode)obj);

                case DynProgramStatementNode.NodeKind.If:
                    return GetHashCode((DynProgramIfNode)obj);

                case DynProgramStatementNode.NodeKind.While:
                    return GetHashCode((DynProgramWhileNode)obj);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return GetHashCode((DynProgramAssignmentNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramContractDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramContractDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramExpressionNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramExpressionNode obj)
        {
            switch (obj.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    return GetHashCode((DynProgramMethodCallNode)obj);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return GetHashCode((DynProgramConstructorCallNode)obj);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynProgramVariableNode)obj);

                case DynProgramStatementNode.NodeKind.If:
                    return GetHashCode((DynProgramIfNode)obj);

                case DynProgramStatementNode.NodeKind.While:
                    return GetHashCode((DynProgramWhileNode)obj);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return GetHashCode((DynProgramAssignmentNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramMethodCallNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramMethodCallNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramConstructorCallNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramConstructorCallNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramIfNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramIfNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.FalseBranch, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.TrueBranch, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramWhileNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramWhileNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Body, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Invariant);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramAssignmentNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramAssignmentNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                result = result * HASH_PRIME + GetHashCode(obj.Variable);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynHoareNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynHoareNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + GetHashCode(obj.PostCondition);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Statements, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.PreCondition);
                result = result * HASH_PRIME + obj.Location.GetHashCode();
                return result;
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionExpressionNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionExpressionNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionExpressionNode x, DynAssertionExpressionNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return Equals((DynAssertionQuantifierNode)x, (DynAssertionQuantifierNode)y);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return Equals((DynAssertionTypeCheckNode)x, (DynAssertionTypeCheckNode)y);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return Equals((DynAssertionPredicateNode)x, (DynAssertionPredicateNode)y);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return Equals((DynAssertionUnaryNode)x, (DynAssertionUnaryNode)y);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    return Equals((DynAssertionBinaryNode)x, (DynAssertionBinaryNode)y);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return Equals((DynAssertionIfThenElseNode)x, (DynAssertionIfThenElseNode)y);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return Equals((DynAssertionNumericNode)x, (DynAssertionNumericNode)y);

                case DynAssertionExpressionNode.NodeKind.String:
                    return Equals((DynAssertionStringNode)x, (DynAssertionStringNode)y);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return Equals((DynAssertionInstanceVariableNode)x, (DynAssertionInstanceVariableNode)y);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return Equals((DynAssertionVariableNode)x, (DynAssertionVariableNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionQuantifierNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionQuantifierNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionQuantifierNode x, DynAssertionQuantifierNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.Expression, y.Expression) &&
                    STRING_COMPARER.Equals(x.Type, y.Type) &&
                    STRING_COMPARER.Equals(x.Variable, y.Variable);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionTypeCheckNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionTypeCheckNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionTypeCheckNode x, DynAssertionTypeCheckNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.TypeSet, y.TypeSet, Equals) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionPredicateNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionPredicateNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionPredicateNode x, DynAssertionPredicateNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionUnaryNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionUnaryNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionUnaryNode x, DynAssertionUnaryNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionBinaryNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionBinaryNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionBinaryNode x, DynAssertionBinaryNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.Second, y.Second) &&
                    Equals(x.First, y.First);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionIfThenElseNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionIfThenElseNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionIfThenElseNode x, DynAssertionIfThenElseNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.FalseBranch, y.FalseBranch) &&
                    Equals(x.TrueBranch, y.TrueBranch) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionNumericNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionNumericNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionNumericNode x, DynAssertionNumericNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    x.Value == y.Value;
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionStringNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionStringNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionStringNode x, DynAssertionStringNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    STRING_COMPARER.Equals(x.Value, y.Value);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionInstanceVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionInstanceVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionInstanceVariableNode x, DynAssertionInstanceVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionVariableNode x, DynAssertionVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramClassNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramClassNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramClassNode x, DynProgramClassNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Contracts, y.Contracts, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Methods, y.Methods, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Renames, y.Renames, Equals) &&
                    STRING_COMPARER.Equals(x.Parent, y.Parent) &&
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    x.Location == y.Location;
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramRenameDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramRenameDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramRenameDeclarationNode x, DynProgramRenameDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    STRING_COMPARER.Equals(x.To, y.To) &&
                    STRING_COMPARER.Equals(x.From, y.From) &&
                    x.Location == y.Location;
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramMethodDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramMethodDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramMethodDeclarationNode x, DynProgramMethodDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Contracts, y.Contracts, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Statements, y.Statements, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Parameters, y.Parameters, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    x.Location == y.Location;
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramStatementNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramStatementNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramStatementNode x, DynProgramStatementNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return Equals((DynProgramContractDeclarationNode)x, (DynProgramContractDeclarationNode)y);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Equals((DynProgramMethodCallNode)x, (DynProgramMethodCallNode)y);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Equals((DynProgramConstructorCallNode)x, (DynProgramConstructorCallNode)y);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Equals((DynProgramVariableNode)x, (DynProgramVariableNode)y);

                case DynProgramStatementNode.NodeKind.If:
                    return Equals((DynProgramIfNode)x, (DynProgramIfNode)y);

                case DynProgramStatementNode.NodeKind.While:
                    return Equals((DynProgramWhileNode)x, (DynProgramWhileNode)y);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Equals((DynProgramAssignmentNode)x, (DynProgramAssignmentNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramContractDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramContractDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramContractDeclarationNode x, DynProgramContractDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramExpressionNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramExpressionNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramExpressionNode x, DynProgramExpressionNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Equals((DynProgramMethodCallNode)x, (DynProgramMethodCallNode)y);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Equals((DynProgramConstructorCallNode)x, (DynProgramConstructorCallNode)y);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Equals((DynProgramVariableNode)x, (DynProgramVariableNode)y);

                case DynProgramStatementNode.NodeKind.If:
                    return Equals((DynProgramIfNode)x, (DynProgramIfNode)y);

                case DynProgramStatementNode.NodeKind.While:
                    return Equals((DynProgramWhileNode)x, (DynProgramWhileNode)y);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Equals((DynProgramAssignmentNode)x, (DynProgramAssignmentNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramMethodCallNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramMethodCallNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramMethodCallNode x, DynProgramMethodCallNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramConstructorCallNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramConstructorCallNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramConstructorCallNode x, DynProgramConstructorCallNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramVariableNode x, DynProgramVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramIfNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramIfNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramIfNode x, DynProgramIfNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.FalseBranch, y.FalseBranch, Equals) &&
                    Utilities.EnumerableDeepEquals(x.TrueBranch, y.TrueBranch, Equals) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramWhileNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramWhileNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramWhileNode x, DynProgramWhileNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Utilities.EnumerableDeepEquals(x.Body, y.Body, Equals) &&
                    Equals(x.Invariant, y.Invariant) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramAssignmentNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramAssignmentNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramAssignmentNode x, DynProgramAssignmentNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Location == y.Location &&
                    Equals(x.Expression, y.Expression) &&
                    Equals(x.Variable, y.Variable);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynHoareNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynHoareNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynHoareNode x, DynHoareNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Equals(x.PostCondition, y.PostCondition) &&
                    Utilities.EnumerableDeepEquals(x.Statements, y.Statements, Equals) &&
                    Equals(x.PreCondition, y.PreCondition) &&
                    x.Location == y.Location;
        }
    }

    //
    // EqualityComparer without Location
    //
    //
    public sealed class DynNodeIgnoreLocationEqualityComparer :
                    IEqualityComparer<DynAssertionExpressionNode>,
                    IEqualityComparer<DynAssertionQuantifierNode>,
                    IEqualityComparer<DynAssertionTypeCheckNode>,
                    IEqualityComparer<DynAssertionPredicateNode>,
                    IEqualityComparer<DynAssertionUnaryNode>,
                    IEqualityComparer<DynAssertionBinaryNode>,
                    IEqualityComparer<DynAssertionIfThenElseNode>,
                    IEqualityComparer<DynAssertionNumericNode>,
                    IEqualityComparer<DynAssertionStringNode>,
                    IEqualityComparer<DynAssertionInstanceVariableNode>,
                    IEqualityComparer<DynAssertionVariableNode>,
                    IEqualityComparer<DynProgramClassNode>,
                    IEqualityComparer<DynProgramRenameDeclarationNode>,
                    IEqualityComparer<DynProgramMethodDeclarationNode>,
                    IEqualityComparer<DynProgramStatementNode>,
                    IEqualityComparer<DynProgramContractDeclarationNode>,
                    IEqualityComparer<DynProgramExpressionNode>,
                    IEqualityComparer<DynProgramMethodCallNode>,
                    IEqualityComparer<DynProgramConstructorCallNode>,
                    IEqualityComparer<DynProgramVariableNode>,
                    IEqualityComparer<DynProgramIfNode>,
                    IEqualityComparer<DynProgramWhileNode>,
                    IEqualityComparer<DynProgramAssignmentNode>,
                    IEqualityComparer<DynHoareNode>
    {
        private static readonly IEqualityComparer<string> STRING_COMPARER = StringComparer.Ordinal;
        private const int HASH_PRIME = Utilities.HASH_PRIME;
        private const int HASH_SEED = Utilities.HASH_SEED;

        private static int GetHashCode(string value)
        {
            return value != null ? value.GetHashCode() : 0;
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionExpressionNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionExpressionNode obj)
        {
            switch (obj.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return GetHashCode((DynAssertionQuantifierNode)obj);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return GetHashCode((DynAssertionTypeCheckNode)obj);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return GetHashCode((DynAssertionPredicateNode)obj);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return GetHashCode((DynAssertionUnaryNode)obj);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    return GetHashCode((DynAssertionBinaryNode)obj);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return GetHashCode((DynAssertionIfThenElseNode)obj);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return GetHashCode((DynAssertionNumericNode)obj);

                case DynAssertionExpressionNode.NodeKind.String:
                    return GetHashCode((DynAssertionStringNode)obj);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynAssertionInstanceVariableNode)obj);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return GetHashCode((DynAssertionVariableNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionQuantifierNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionQuantifierNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                result = result * HASH_PRIME + GetHashCode(obj.Type);
                result = result * HASH_PRIME + GetHashCode(obj.Variable);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionTypeCheckNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionTypeCheckNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.TypeSet, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionPredicateNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionPredicateNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionUnaryNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionUnaryNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionBinaryNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionBinaryNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Second);
                result = result * HASH_PRIME + GetHashCode(obj.First);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionIfThenElseNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionIfThenElseNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.FalseBranch);
                result = result * HASH_PRIME + GetHashCode(obj.TrueBranch);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionNumericNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionNumericNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + obj.Value.GetHashCode();
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionStringNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionStringNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Value);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionInstanceVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionInstanceVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynAssertionVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynAssertionVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramClassNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramClassNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Contracts, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Methods, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Renames, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Parent);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramRenameDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramRenameDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + GetHashCode(obj.To);
                result = result * HASH_PRIME + GetHashCode(obj.From);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramMethodDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramMethodDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Contracts, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Statements, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Parameters, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramStatementNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramStatementNode obj)
        {
            switch (obj.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return GetHashCode((DynProgramContractDeclarationNode)obj);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return GetHashCode((DynProgramMethodCallNode)obj);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return GetHashCode((DynProgramConstructorCallNode)obj);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynProgramVariableNode)obj);

                case DynProgramStatementNode.NodeKind.If:
                    return GetHashCode((DynProgramIfNode)obj);

                case DynProgramStatementNode.NodeKind.While:
                    return GetHashCode((DynProgramWhileNode)obj);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return GetHashCode((DynProgramAssignmentNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramContractDeclarationNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramContractDeclarationNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramExpressionNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramExpressionNode obj)
        {
            switch (obj.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    return GetHashCode((DynProgramMethodCallNode)obj);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return GetHashCode((DynProgramConstructorCallNode)obj);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return GetHashCode((DynProgramVariableNode)obj);

                case DynProgramStatementNode.NodeKind.If:
                    return GetHashCode((DynProgramIfNode)obj);

                case DynProgramStatementNode.NodeKind.While:
                    return GetHashCode((DynProgramWhileNode)obj);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return GetHashCode((DynProgramAssignmentNode)obj);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("obj.Kind", (int)obj.Kind);
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramMethodCallNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramMethodCallNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramConstructorCallNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramConstructorCallNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Arguments, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramVariableNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramVariableNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Name);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramIfNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramIfNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.FalseBranch, GetHashCode);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.TrueBranch, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramWhileNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramWhileNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Body, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.Invariant);
                result = result * HASH_PRIME + GetHashCode(obj.Condition);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynProgramAssignmentNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynProgramAssignmentNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + obj.Kind.GetHashCode();
                result = result * HASH_PRIME + GetHashCode(obj.Expression);
                result = result * HASH_PRIME + GetHashCode(obj.Variable);
                return result;
            }
        }

        /// <summary>Returns a hash code for the specified object.</summary>
        /// <param name="obj">The <see cref="DynHoareNode" /> for which a hash code is to be returned.</param>
        /// <returns>A hash code for the specified object.</returns>
        public int GetHashCode(DynHoareNode obj)
        {
            unchecked
            {
                int result = HASH_SEED;
                result = result * HASH_PRIME + GetHashCode(obj.PostCondition);
                result = result * HASH_PRIME + Utilities.EnumerableDeepGetHashCode(obj.Statements, GetHashCode);
                result = result * HASH_PRIME + GetHashCode(obj.PreCondition);
                return result;
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionExpressionNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionExpressionNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionExpressionNode x, DynAssertionExpressionNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.Universal:
                case DynAssertionExpressionNode.NodeKind.Existential:
                    return Equals((DynAssertionQuantifierNode)x, (DynAssertionQuantifierNode)y);

                case DynAssertionExpressionNode.NodeKind.TypeCheck:
                    return Equals((DynAssertionTypeCheckNode)x, (DynAssertionTypeCheckNode)y);

                case DynAssertionExpressionNode.NodeKind.Predicate:
                    return Equals((DynAssertionPredicateNode)x, (DynAssertionPredicateNode)y);

                case DynAssertionExpressionNode.NodeKind.Negate:
                    return Equals((DynAssertionUnaryNode)x, (DynAssertionUnaryNode)y);

                case DynAssertionExpressionNode.NodeKind.And:
                case DynAssertionExpressionNode.NodeKind.Or:
                case DynAssertionExpressionNode.NodeKind.Implies:
                case DynAssertionExpressionNode.NodeKind.Addition:
                case DynAssertionExpressionNode.NodeKind.Subtraction:
                case DynAssertionExpressionNode.NodeKind.Multiplication:
                case DynAssertionExpressionNode.NodeKind.Division:
                case DynAssertionExpressionNode.NodeKind.Modulus:
                case DynAssertionExpressionNode.NodeKind.GreaterThan:
                case DynAssertionExpressionNode.NodeKind.GreaterThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.LessThan:
                case DynAssertionExpressionNode.NodeKind.LessThanOrEqual:
                case DynAssertionExpressionNode.NodeKind.Equality:
                    return Equals((DynAssertionBinaryNode)x, (DynAssertionBinaryNode)y);

                case DynAssertionExpressionNode.NodeKind.IfThenElse:
                    return Equals((DynAssertionIfThenElseNode)x, (DynAssertionIfThenElseNode)y);

                case DynAssertionExpressionNode.NodeKind.Numeric:
                    return Equals((DynAssertionNumericNode)x, (DynAssertionNumericNode)y);

                case DynAssertionExpressionNode.NodeKind.String:
                    return Equals((DynAssertionStringNode)x, (DynAssertionStringNode)y);

                case DynAssertionExpressionNode.NodeKind.InstanceVariable:
                    return Equals((DynAssertionInstanceVariableNode)x, (DynAssertionInstanceVariableNode)y);

                case DynAssertionExpressionNode.NodeKind.AssertionVariable:
                case DynAssertionExpressionNode.NodeKind.ProgramVariable:
                    return Equals((DynAssertionVariableNode)x, (DynAssertionVariableNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynAssertionExpressionNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionQuantifierNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionQuantifierNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionQuantifierNode x, DynAssertionQuantifierNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.Expression, y.Expression) &&
                    STRING_COMPARER.Equals(x.Type, y.Type) &&
                    STRING_COMPARER.Equals(x.Variable, y.Variable);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionTypeCheckNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionTypeCheckNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionTypeCheckNode x, DynAssertionTypeCheckNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.TypeSet, y.TypeSet, Equals) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionPredicateNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionPredicateNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionPredicateNode x, DynAssertionPredicateNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionUnaryNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionUnaryNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionUnaryNode x, DynAssertionUnaryNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionBinaryNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionBinaryNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionBinaryNode x, DynAssertionBinaryNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.Second, y.Second) &&
                    Equals(x.First, y.First);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionIfThenElseNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionIfThenElseNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionIfThenElseNode x, DynAssertionIfThenElseNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.FalseBranch, y.FalseBranch) &&
                    Equals(x.TrueBranch, y.TrueBranch) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionNumericNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionNumericNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionNumericNode x, DynAssertionNumericNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    x.Value == y.Value;
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionStringNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionStringNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionStringNode x, DynAssertionStringNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    STRING_COMPARER.Equals(x.Value, y.Value);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionInstanceVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionInstanceVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionInstanceVariableNode x, DynAssertionInstanceVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynAssertionVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynAssertionVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynAssertionVariableNode x, DynAssertionVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramClassNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramClassNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramClassNode x, DynProgramClassNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Contracts, y.Contracts, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Methods, y.Methods, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Renames, y.Renames, Equals) &&
                    STRING_COMPARER.Equals(x.Parent, y.Parent) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramRenameDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramRenameDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramRenameDeclarationNode x, DynProgramRenameDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    STRING_COMPARER.Equals(x.To, y.To) &&
                    STRING_COMPARER.Equals(x.From, y.From);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramMethodDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramMethodDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramMethodDeclarationNode x, DynProgramMethodDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Contracts, y.Contracts, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Statements, y.Statements, Equals) &&
                    Utilities.EnumerableDeepEquals(x.Parameters, y.Parameters, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramStatementNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramStatementNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramStatementNode x, DynProgramStatementNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynProgramStatementNode.NodeKind.Ensures:
                case DynProgramStatementNode.NodeKind.Requires:
                case DynProgramStatementNode.NodeKind.Invariant:
                case DynProgramStatementNode.NodeKind.Assert:
                case DynProgramStatementNode.NodeKind.ByTheWay:
                    return Equals((DynProgramContractDeclarationNode)x, (DynProgramContractDeclarationNode)y);

                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Equals((DynProgramMethodCallNode)x, (DynProgramMethodCallNode)y);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Equals((DynProgramConstructorCallNode)x, (DynProgramConstructorCallNode)y);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Equals((DynProgramVariableNode)x, (DynProgramVariableNode)y);

                case DynProgramStatementNode.NodeKind.If:
                    return Equals((DynProgramIfNode)x, (DynProgramIfNode)y);

                case DynProgramStatementNode.NodeKind.While:
                    return Equals((DynProgramWhileNode)x, (DynProgramWhileNode)y);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Equals((DynProgramAssignmentNode)x, (DynProgramAssignmentNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramContractDeclarationNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramContractDeclarationNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramContractDeclarationNode x, DynProgramContractDeclarationNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramExpressionNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramExpressionNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramExpressionNode x, DynProgramExpressionNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            switch (x.Kind)
            {
                case DynProgramStatementNode.NodeKind.MethodCall:
                    return Equals((DynProgramMethodCallNode)x, (DynProgramMethodCallNode)y);

                case DynProgramStatementNode.NodeKind.ConstructorCall:
                    return Equals((DynProgramConstructorCallNode)x, (DynProgramConstructorCallNode)y);

                case DynProgramStatementNode.NodeKind.Variable:
                case DynProgramStatementNode.NodeKind.InstanceVariable:
                    return Equals((DynProgramVariableNode)x, (DynProgramVariableNode)y);

                case DynProgramStatementNode.NodeKind.If:
                    return Equals((DynProgramIfNode)x, (DynProgramIfNode)y);

                case DynProgramStatementNode.NodeKind.While:
                    return Equals((DynProgramWhileNode)x, (DynProgramWhileNode)y);

                case DynProgramStatementNode.NodeKind.Assignment:
                    return Equals((DynProgramAssignmentNode)x, (DynProgramAssignmentNode)y);

                default:
                    throw Guard.EnumOutOfRangeException<DynProgramStatementNode.NodeKind>("x.Kind", (int)x.Kind);
            }
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramMethodCallNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramMethodCallNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramMethodCallNode x, DynProgramMethodCallNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name) &&
                    Equals(x.Expression, y.Expression);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramConstructorCallNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramConstructorCallNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramConstructorCallNode x, DynProgramConstructorCallNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Arguments, y.Arguments, Equals) &&
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramVariableNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramVariableNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramVariableNode x, DynProgramVariableNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    STRING_COMPARER.Equals(x.Name, y.Name);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramIfNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramIfNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramIfNode x, DynProgramIfNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.FalseBranch, y.FalseBranch, Equals) &&
                    Utilities.EnumerableDeepEquals(x.TrueBranch, y.TrueBranch, Equals) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramWhileNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramWhileNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramWhileNode x, DynProgramWhileNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Utilities.EnumerableDeepEquals(x.Body, y.Body, Equals) &&
                    Equals(x.Invariant, y.Invariant) &&
                    Equals(x.Condition, y.Condition);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynProgramAssignmentNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynProgramAssignmentNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynProgramAssignmentNode x, DynProgramAssignmentNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }
            if (x.Kind != y.Kind) { return false; }

            return
                    Equals(x.Expression, y.Expression) &&
                    Equals(x.Variable, y.Variable);
        }

        /// <summary>Determines whether the specified objects are equal.</summary>
        /// <param name="x">The first object of type <see cref="DynHoareNode" /> to compare.</param>
        /// <param name="y">The second object of type <see cref="DynHoareNode" /> to compare.</param>
        /// <returns><code>true</code> if the specified objects are equal; otherwise, <code>false</code>.</returns>
        public bool Equals(DynHoareNode x, DynHoareNode y)
        {
            if (object.ReferenceEquals(x, y)) { return true; }
            if (object.ReferenceEquals(x, null) | object.ReferenceEquals(y, null)) { return false; }

            return
                    Equals(x.PostCondition, y.PostCondition) &&
                    Utilities.EnumerableDeepEquals(x.Statements, y.Statements, Equals) &&
                    Equals(x.PreCondition, y.PreCondition);
        }
    }
}
