﻿using System;
using System.Collections.Generic;

namespace Dyn
{
#if DEBUG
    [System.Diagnostics.DebuggerStepThrough]
#endif
    public static class Guard
    {
        public static Exception EnumOutOfRangeException<T>(string paramName, int value)
        {
            EnsureEnumDefined<T>(paramName, value);
            return new ArgumentException(paramName + " has invalid value " + Enum.GetName(typeof(T), value) + ".", paramName);
        }

        public static void EnsureNotNull<T>(string paramName, T value)
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName);
            }
        }

        public static void EnsureElementNotNull<T>(string paramName, IEnumerable<T> value) where T : class
        {
            int counter = 0;
            foreach (var element in value)
            {
                EnsureNotNull<T>(paramName + "[" + counter + "]", element);
                ++counter;
            }
        }

        public static void EnsureEnumDefined<T>(string paramName, int value)
        {
            if (!Enum.IsDefined(typeof(T), value))
            {
                throw new ArgumentOutOfRangeException(paramName, "value not defined in Enum"); //TODO: Localize
            }
        }

        public static void EnsureNonnegative(string paramName, int value)
        {
            if (value < 0)
            {
                throw new ArgumentOutOfRangeException(paramName, "value is negative"); //TODO: Localize
            }
        }
    }
}
