﻿using Dyn.Lexer;
using System.Collections.Generic;

namespace Dyn.Parser
{
    public static class Compatibility
    {
        public static IEnumerable<SyntaxToken> TransformCommentToAssertion(this IEnumerable<SyntaxToken> source)
        {
            foreach (var element in source)
            {
                if (element.Kind != SyntaxKind.CommentKind)
                {
                    yield return element;
                    continue;
                }

                var comment = element.String.Trim();

                if (!comment.StartsWith("#{") || !comment.EndsWith("};"))
                {
                    yield return element;
                    continue;
                }

                var originalSpan = element.Span;

                var assertion = comment.Substring(2, comment.Length - 4);

                yield return new SyntaxToken(
                    SyntaxKind.AssertKeyword,
                    new Span(
                        originalSpan.FileName,
                        originalSpan.Position,
                        originalSpan.Line,
                        originalSpan.Column,
                        2));

                foreach (var token in new DynLexer(assertion))
                {
                    var currentSpan = token.Span;

                    yield return token.SetSpan(new Span(
                        originalSpan.FileName,
                        originalSpan.Position + 2 + currentSpan.Position,
                        originalSpan.Line + currentSpan.Line,
                        originalSpan.Column + 2 + currentSpan.Column,
                        currentSpan.Length));
                }
            }
        }

        public static IEnumerable<SyntaxToken> AddWhileInvariant(this IEnumerable<SyntaxToken> source)
        {
            foreach (var element in source)
            {
                if (element.Kind == SyntaxKind.DoKeyword)
                {
                    yield return new SyntaxToken(SyntaxKind.InvariantKeyword, Span.Empty);
                    yield return new SyntaxToken(SyntaxKind.TrueKeyword, Span.Empty);
                }

                yield return element;
            }
        }
    }
}
