﻿using System;
using System.IO;

namespace Dyn.Lexer
{
    internal sealed class SourceReaderHelper
    {
        private readonly TextReader reader;
        private readonly string fileName;

        private const int BUFFER = 1024; // 1 kB buffer

        private char[] buffer = new char[BUFFER];
        private int streamBufferStart = 0;
        private int bufferCursorPosition = 0;
        private int lineCount = 0;
        private int lastLinebreakPosition = 0;
        private int max = 0;
        public char Current = default(char);

        public SourceReaderHelper(String fileName, TextReader reader)
        {
            Guard.EnsureNotNull("fileName", fileName);
            Guard.EnsureNotNull("reader", reader);

            this.reader = reader;
            this.fileName = fileName;
        }

        public string FileName
        {
            get
            {
                return this.fileName;
            }
        }

        // TODO: Fix Name
        public bool BeforeContent
        {
            get
            {
                return this.streamBufferStart + this.bufferCursorPosition == 0;
            }
        }

        public bool AtContent
        {
            get
            {
                return this.bufferCursorPosition <= max;
            }
        }

        public bool AfterContent
        {
            get
            {
                return this.bufferCursorPosition > max;
            }
        }

        public Span CreateSpan()
        {
            if (this.bufferCursorPosition == 0 && this.streamBufferStart == 0)
            {
                return new Span(FileName, 0, 0, 0, 0);
            }

            var pos = this.streamBufferStart + this.bufferCursorPosition - 1;

            // Position is the number of already read charaters and not as in Span the index of the next character
            return new Span(FileName, pos, this.lineCount, pos - this.lastLinebreakPosition, 0);
        }

        public bool TryNext()
        {
            // end of stream reached
            if (bufferCursorPosition > max)
            {
                Current = default(char);
                return false;
            }

            // handle previous new lines to make sure that the new line character is the last character of that line
            if (Current == '\n')
            {
                lineCount++;
                lastLinebreakPosition = bufferCursorPosition + streamBufferStart;
            }

            // end of buffer reached
            if (bufferCursorPosition == max)
            {
                int result = reader.Read(buffer, 0, buffer.Length);

                if (result <= 0)
                {
                    ++bufferCursorPosition;
                    Current = default(char);
                    return false;
                }

                streamBufferStart += max;
                bufferCursorPosition = 0;
                max = result;
            }

            Current = buffer[bufferCursorPosition++];

            return true;
        }
    }
}
