﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Dyn.Lexer
{
    public static class SyntaxTokenExtensions
    {
        public static SyntaxToken SetSpan(this SyntaxToken source, Span value)
        {
            switch (source.Kind)
            {
                case SyntaxKind.NumericKind:
                    return new SyntaxToken(source.Kind, value, source.Number);

                case SyntaxKind.InvalidKind:
                case SyntaxKind.WhitespaceKind:
                case SyntaxKind.StringKind:
                case SyntaxKind.CommentKind:
                case SyntaxKind.IdentifierKind:
                    return new SyntaxToken(source.Kind, value, source.String);

                default:
                    return new SyntaxToken(source.Kind, value);
            }
        }

        public static IEnumerable<SyntaxToken> ExtractSpan(this IEnumerable<SyntaxToken> source, Span region)
        {
            // TODO: More complex system:
            // Splits up into lines

            return source.Where(e => InRange(region, e.Span));
        }

        /*
        private static void WriteCodeSegment(IEnumerable<SyntaxToken> source, Span location)
        {
            SyntaxToken previous = default(SyntaxToken);
            bool isPreviousValid = false;

            var result = new List<SyntaxToken>();

            foreach (var token in source)
            {
                var span = token.Span;
                isPreviousValid = false;

                if (!StringComparer.Ordinal.Equals(span.FileName, location.FileName))
                {
                    previous = token;
                    continue;
                }

                if(span.Line < location.Line)
                {
                    previous = token;
                    continue;
                }

                if(span.Line == location.Line)
                {

                    if(previous.Span.Position + previous.Span.Length > span.Position - span.Column)
                    {
                        var lastLineOfPreviousSpan = new Span(
                            previous.Span.FileName,
                            span.Position - span.Column,
                            span.Line,


                            );

                        var lastLineOfPrevious = new SyntaxToken(
                            previous.Kind,
                            previous.

                            )

                        result.Add(lastLineOfPrevious);
                    }






                }



            }

            if (isPreviousValid)
            {
                result.Add(previous);

                ColoredFormatter.PrintTokenListHighlighted(new SyntaxToken[1] { previous });
            }





            ColoredFormatter.PrintTokenListHighlighted(
                this._tokens.Where(e => InRange(location, e.Span)));


            throw new NotImplementedException();
        }
*/

        private static bool InRange(Span reference, Span other)
        {
            // same file
            if (!StringComparer.Ordinal.Equals(reference.FileName, other.FileName))
            {
                return false;
            }

            int start = reference.Position;
            int end = reference.Position + reference.Length;

            // starts in the same line
            if (reference.Line == other.Line)
            {
                return true;
            }

            // not the same line but directly preceding
            if (reference.Position == reference.Position + reference.Length)
            {
                return true;
            }

            return false;
        }
    }
}
