﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Dyn.SmtLib2
{
    public sealed class SexprWriter
    {
        private readonly TextWriter _writer;
        private readonly bool _indent;

        private const string DEFAULT_INDENTATION = "  ";
        private int depth = 0;
        private bool tail = false;

        public SexprWriter(TextWriter writer, bool indent)
        {
            Guard.EnsureNotNull("writer", writer);

            this._writer = writer;
            this._indent = indent;
        }

        private void Indent(bool close)
        {
            // no new line at the beginning
            if (!tail)
            {
                return;
            }

            if (!this._indent)
            {
                if(!close)
                {
                    this._writer.Write(' ');
                }
                return;
            }

            this._writer.WriteLine();
            for (int i = 0; i < depth; ++i)
            {
                this._writer.Write(DEFAULT_INDENTATION);
            }
        }

        public SexprWriter Open(string atom)
        {
            Guard.EnsureNotNull("atom", atom);

            Indent(false);

            this._writer.Write("(");
            this._writer.Write(atom);
            ++depth;

            tail = true;

            return this;
        }

        public SexprWriter Symbol(string atom)
        {
            Indent(false);

            this._writer.Write(atom);

            tail = true;

            return this;
        }

        public SexprWriter Close()
        {
            if (depth <= 0)
            {
                // empty exception
                throw new IndexOutOfRangeException(); //TODO: Exception language
            }

            --depth;

            Indent(true);

            this._writer.Write(")");

            return this;
        }
    }
}
