﻿using System;
using System.Collections.Generic;

namespace Dyn
{
    public static class Utilities
    {
        private sealed class EmptyList<T> : IList<T>, IEnumerator<T>
        {
            private static void ThrowImmutable()
            {
                throw new InvalidOperationException("Cannot change read only instance");
            }

            #region IList<T> Members

            public int IndexOf(T item)
            {
                return -1;
            }

            public void Insert(int index, T item)
            {
                ThrowImmutable();
            }

            public void RemoveAt(int index)
            {
                ThrowImmutable();
            }

            public T this[int index]
            {
                get
                {
                    throw new ArgumentOutOfRangeException("index");
                }
                set
                {
                    throw new ArgumentOutOfRangeException("index");
                }
            }

            #endregion IList<T> Members

            #region ICollection<T> Members

            public void Add(T item)
            {
                ThrowImmutable();
            }

            public void Clear()
            {
                ThrowImmutable();
            }

            public bool Contains(T item)
            {
                return false;
            }

            public void CopyTo(T[] array, int arrayIndex)
            {
                if (array == null)
                {
                    throw new ArgumentNullException("array");
                }
            }

            public int Count
            {
                get { return 0; }
            }

            public bool IsReadOnly
            {
                get { return true; }
            }

            public bool Remove(T item)
            {
                return false;
            }

            #endregion ICollection<T> Members

            #region IEnumerable<T> Members

            public IEnumerator<T> GetEnumerator()
            {
                return this;
            }

            System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
            {
                return ((IEnumerable<T>)this).GetEnumerator();
            }

            #endregion IEnumerable<T> Members

            #region IEnumerator Members

            public T Current
            {
                get { throw new InvalidOperationException(); }
            }

            public void Dispose()
            {
            }

            object System.Collections.IEnumerator.Current
            {
                get { return ((IEnumerator<T>)this).Current; }
            }

            public bool MoveNext()
            {
                return false;
            }

            public void Reset()
            {
            }

            #endregion IEnumerator Members
        }

        public static IList<T> Empty<T>()
        {
            // TODO: static variable
            return new EmptyList<T>();
        }

        [System.Diagnostics.DebuggerHidden]
        public static T[] AsArray<T>(params T[] source)
        {
            return source;
        }
    }
}
