﻿using Dyn.Ast;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Dyn.VerficationConditions
{
    public class SimplifierTransform : DynAssertionNodeTransformer
    {
        public bool IsTrue(DynAssertionExpressionNode assertion) {
            return AstExtensions.IgnoreLocation.Equals(assertion, AstExtensions.Assertions.True);
        }

        public bool IsFalse(DynAssertionExpressionNode assertion) {
          return AstExtensions.IgnoreLocation.Equals(assertion, AstExtensions.Assertions.False);
        }

        public override DynAssertionExpressionNode Transform(DynAssertionBinaryNode node)
        {
            DynAssertionExpressionNode newFirst = null;
            DynAssertionExpressionNode newSecond = null;
            switch (node.Kind)
            {
                case DynAssertionExpressionNode.NodeKind.And:
                  newFirst = Transform(node.First);
                  newSecond = Transform(node.Second);
                  if (IsFalse(newFirst) || IsFalse(newSecond)) {
                    return AstExtensions.Assertions.False;
                  }
                  if (IsTrue(newFirst) && IsTrue(newSecond)) {
                    return AstExtensions.Assertions.True;
                  }
                  if (IsTrue(newFirst)) {
                    return newSecond;
                  }
                  if (IsTrue(newSecond)) {
                    return newFirst;
                  }
                  return new DynAssertionBinaryNode(node.Location, node.Kind, newFirst, newSecond);

                case DynAssertionExpressionNode.NodeKind.Or:
                  newFirst = Transform(node.First);
                  newSecond = Transform(node.Second);
                  if (IsTrue(newFirst) || IsTrue(newSecond)) {
                    return AstExtensions.Assertions.True;
                  }
                  if (IsFalse(newFirst) && IsFalse(newSecond)) {
                    return AstExtensions.Assertions.False;
                  }
                  if (IsFalse(newFirst)) {
                    return newSecond;
                  }
                  if (IsFalse(newSecond)) {
                    return newFirst;
                  }
                  return new DynAssertionBinaryNode(node.Location, node.Kind, newFirst, newSecond);

                case DynAssertionExpressionNode.NodeKind.Implies:
                  newFirst = Transform(node.First);
                  newSecond = Transform(node.Second);
                  if (IsFalse(newFirst) || IsTrue(newSecond)) {
                    return AstExtensions.Assertions.True;
                  }
                  if (IsTrue(newFirst)) {
                    return newSecond;
                  }
                  if (IsFalse(newSecond)) {
                    // ! newFirst
                    return new DynAssertionUnaryNode(node.Location, newFirst);
                  }
                  return new DynAssertionBinaryNode(node.Location, node.Kind, newFirst, newSecond);

                default:
                  return base.Transform(node);
            }
        }

        public override DynAssertionExpressionNode Transform(DynAssertionQuantifierNode node)
        {
            DynAssertionExpressionNode newExp = Transform(node.Expression);
            if (!newExp.FreeVariables().Contains(string.Format("'{0}", node.Variable))) {
              return newExp;
            }
            return new DynAssertionQuantifierNode(node.Location, node.Kind, node.Variable, node.Type, newExp);
        }

    }

    public static class Simplifier
    {
        //TODO: singleton
        public static DynAssertionExpressionNode Simplify(DynAssertionExpressionNode assertion)
        {
          SimplifierTransform transform = new SimplifierTransform();
          return transform.Transform(assertion);
        }
    }

}

