﻿using System;
using System.Collections.Generic;

namespace libdyn.generator
{
#if DEBUG

    [System.Diagnostics.DebuggerStepThrough]
#endif
    public static class Guard
    {
        public static void EnsureNotNull<T>(string paramName, T value) where T : class
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName);
            }
        }

        public static void EnsureElementNotNull<T>(string paramName, IEnumerable<T> value) where T : class
        {
            int counter = 0;
            foreach (var element in value)
            {
                EnsureNotNull<T>(paramName + "[" + counter + "]", element);
                ++counter;
            }
        }

        public static void EnsureEnumDefined<T>(string paramName, int value)
        {
            if (!Enum.IsDefined(typeof(T), value))
            {
                throw new ArgumentOutOfRangeException(paramName, "value not defined in Enum"); //TODO: Localize
            }
        }
    }
}
