﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace libdyn.generator.Lexer
{
    public static class DfaExtensions
    {
        // assuming minimized
        public static string ToRegularExpression(this Dfa source)
        {
            throw new NotImplementedException();
        }

        public static readonly Func<IEnumerable<char>, string> CHARACTER_COMBINER = e => Combine(e, CombineCharPretty);
        public static readonly Func<IEnumerable<char>, string> INT_COMBINER = e => Combine(e, CombineIntPretty);

        private static string CombineCharPretty(char first, char second, bool tail)
        {
            var result = new StringBuilder();

            result.Append(first.EscapeInput());

            if (first != second)
            {
                if (first < second - 1)
                {
                    result.Append('-');
                }

                result.Append(second.EscapeInput());
            }
            return result.ToString();
        }

        private static string CombineIntPretty(char first, char second, bool tail)
        {
            var result = new StringBuilder();

            if (tail)
            {
                result.Append(", ");
            }

            result.Append((int)first);

            if (first != second)
            {
                if (first < second - 1)
                {
                    result.Append('-');
                }
                else
                {
                    result.Append(", ");
                }

                result.Append((int)second);
            }

            return result.ToString();
        }

        private static string Combine(IEnumerable<char> source, Func<char, char, bool, string> pretty)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("pretty", pretty);

            var arr = source.OrderBy(e => e).ToArray();

            if (arr.Length == 0)
            {
                return string.Empty;
            }

            if (arr.Length == 1)
            {
                return pretty(arr[0], arr[0], false);
            }

            var startChar = arr[0];
            var runLength = 1;

            StringBuilder builder = new StringBuilder();

            bool tail = false;

            for (int i = 1; i <= arr.Length; ++i)
            {
                char current = i < arr.Length ? arr[i] : '\0';

                if (startChar + runLength == current)
                {
                    runLength++;
                }
                else
                {
                    builder.Append(pretty(startChar, arr[i - 1], tail));

                    startChar = current;
                    runLength = 1;

                    tail = true;
                }
            }

            return builder.ToString();
        }

        public static string ToStringCombined(this Dfa source, Func<IEnumerable<char>, string> combiner)
        {
            Guard.EnsureNotNull("source", source);
            Guard.EnsureNotNull("combiner", combiner);

            var builder = new StringBuilder();

            builder.AppendFormat("{0,8}   {1,16} -> {2,8}\n", string.Empty, string.Empty, source.Initial);

            foreach (var grouping in source.Transitions.GroupBy(t => t.Current + "_" + t.Next).OrderBy(e => e.Key, StringComparer.OrdinalIgnoreCase))
            {
                var any = grouping.First();

                builder.AppendFormat("{0,8}   [{1,14}] -> {2,8}\n", any.Current, combiner(grouping.Select(t => t.Input)), any.Next);
            }

            builder.AppendFormat("{0,8}{1,-16}    ->{2,8}\n", string.Empty, string.Join(", ", source.Accepting), string.Empty);

            return builder.ToString();
        }
    }
}
