﻿using System;
using System.Diagnostics;
using input = System.Char;
using state = System.Int32;

namespace libdyn.generator.Lexer
{
    [DebuggerDisplay("{ToString(),nq}")]
    public struct Transition : IEquatable<Transition>
    {
        public const char NOT_FLAG = (char)0x8000;

        public readonly state Current;
        public readonly input Input;
        public readonly state Next;

        public Transition(state from, input value, state to)
        {
            Current = from;
            Next = to;
            Input = value;
        }

        #region IEquatable<Transition> Members

        public bool Equals(Transition other)
        {
            return this.Current == other.Current && this.Next == other.Next && this.Input == other.Input;
        }

        public override int GetHashCode()
        {
            int result = 1;
            result = (result << 5 + result) ^ Current.GetHashCode();
            result = (result << 5 + result) ^ Input.GetHashCode();
            result = (result << 5 + result) ^ Next.GetHashCode();

            return result;
        }

        public override string ToString()
        {
            return string.Format("{0,16} + {1,16} -> {2,16}", Current, Input.EscapeInput(), Next);
        }

        #endregion IEquatable<Transition> Members

        private readonly static Transition[] EMPTY = new Transition[0];

        public static Transition[] Empty
        {
            get
            {
                return EMPTY;
            }
        }
    }
}
