﻿using Dyn;
using System;
using System.Collections;
using System.Collections.Generic;

namespace tests.Ast
{
    public sealed class AstFuzzer : IEnumerable<object>
    {
        private readonly IDictionary<Type, object> _constructors = new Dictionary<Type, object>();

        internal void Add<T>(Func<Random, int, T> constructor)
        {
            Guard.EnsureNotNull("constructor", constructor);
            this._constructors.Add(typeof(T), constructor);
        }

        public TKey Fuzzy<TKey>()
        {
            return Fuzzy<TKey>(new Random(), 0);
        }

        public TKey Fuzzy<TKey>(Random rnd)
        {
            return Fuzzy<TKey>(rnd, 0);
        }

        public TKey Fuzzy<TKey>(Random rnd, int depth)
        {
            Guard.EnsureNotNull("rnd", rnd);
            Guard.EnsureNonnegative("depth", depth);

            if (depth > 200)
            {
                throw new OperationCanceledException();
            }

            object result;

            var key = typeof(TKey);

            if (this._constructors.TryGetValue(key, out result))
            {
                return ((Func<Random, int, TKey>)result)(rnd, depth + 1);
            }

            throw new ArgumentException("Type " + key.ToGenericTypeString() + " not found."); //TODO: description & Localization
        }

        IEnumerator<object> IEnumerable<object>.GetEnumerator()
        {
            throw new NotSupportedException();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new NotSupportedException();
        }
    }
}
