﻿using Dyn;
using System;
using System.Collections.Generic;

namespace tests.Ast
{
    public static class AstFuzzerExtensions
    {
        public static IList<T> List<T>(int count, Random rnd, int depth, Func<Random, int, T> factory)
        {
            var result = new List<T>();

            for (int i = 0; i < count; ++i)
            {
                result.Add(factory(rnd, depth + 1));
            }

            return result;
        }

        public sealed class InheritanceBuilder<TSuper> : PropabilityLookup<Func<Random, int, TSuper>>
        {
            private readonly Func<AstFuzzer> _lasyFuzzer;

            public InheritanceBuilder(Func<AstFuzzer> lazyFuzzer) : base()
            {
                Guard.EnsureNotNull("lazyFuzzer", lazyFuzzer);
                this._lasyFuzzer = lazyFuzzer;
            }

            public void Add<TKey>(TKey ignored, int propability) where TKey : TSuper
            {
                base.Add((r, d) => this._lasyFuzzer().Fuzzy<TKey>(r, d), propability);
            }

            public Func<Random, int, TSuper> ToFunc()
            {
                base.Freeze();
                return (r, d) => base.Get(r)(r, d);
            }
        }

        public sealed class PropabilityBuilder<TSuper> : PropabilityLookup<TSuper>
        {
            public PropabilityBuilder() : base()
            { }

            public new void Add(TSuper key, int propability)
            {
                base.Add(key, propability);
            }

            public Func<Random, int, TSuper> ToFunc()
            {
                return (r, d) => base.Get(r);
            }
        }
    }
}
