﻿using Dyn.Ast;
using System;
using System.Collections.Generic;
using Xunit;

namespace tests.Ast
{
    public static class ComparerTests
    {
        private const string FILE_NAME = "ComparerTests";

        [Theory]
        [InlineData(1, 32)]
        [InlineData(2, 64)]
        [InlineData(3, 128)]
        public static void RandomHashCodeEqualityTest(int seed, int count)
        {
            RandomHashCodeEqualityTestInner<DynAssertionExpressionNode>(seed, count, DynFuzzer.Assertions, DynNodeFormatter.ToSource);
            RandomHashCodeEqualityTestInner<DynProgramClassNode>(seed, count, DynFuzzer.Programs, DynNodeFormatter.ToSource);
            RandomHashCodeEqualityTestInner<DynProgramExpressionNode>(seed, count, DynFuzzer.Programs, DynNodeFormatter.ToSource);
            RandomHashCodeEqualityTestInner<DynProgramStatementNode>(seed, count, DynFuzzer.Programs, DynNodeFormatter.ToSource);
            RandomHashCodeEqualityTestInner<DynHoareNode>(seed, count, DynFuzzer.Programs, DynNodeFormatter.ToSource);
        }

        private static void RandomHashCodeEqualityTestInner<T>(int seed, int count, AstFuzzer fuzzer, Func<T, string> formatter)
        {
            var r = new Random(seed);

            var nec = AstExtensions.Normal as IEqualityComparer<T>;
            var iec = AstExtensions.IgnoreLocation as IEqualityComparer<T>;

            Assert.NotNull(nec);
            Assert.NotNull(iec);

            HashSet<T> nodes1 = new HashSet<T>(nec);
            HashSet<T> nodes2 = new HashSet<T>(iec);
            HashSet<string> nodeSource = new HashSet<string>();

            List<T> nodes0 = new List<T>();
            List<string> nodesSources0 = new List<string>();

            int tcount = count;

            while (tcount > 0)
            {
                var node = fuzzer.Fuzzy<T>(new Random(r.Next()));
                var source = formatter(node);

                bool node1Success = nodes1.Add(node);
                bool node2Success = nodes2.Add(node);
                bool nodeSourceSuccess = nodeSource.Add(source);
                Assert.Equal(node1Success, node2Success);

                Assert.Equal(node1Success, nodeSourceSuccess);

                if (node1Success)
                {
                    nodes0.Add(node);
                    nodesSources0.Add(source);
                    tcount--;
                }
            }

            for (int i = 0; i < count; ++i)
            {
                for (int j = i + 1; j < count; ++j)
                {
                    Assert.NotEqual(nodes0[i], nodes0[j], iec);
                    Assert.NotEqual(nodes0[i], nodes0[j], nec);
                    Assert.NotEqual(nodesSources0[i], nodesSources0[j]);
                }
            }
        }
    }
}
