﻿using Dyn;
using System;
using System.Collections;
using System.Collections.Generic;

namespace tests.Ast
{
    public abstract class PropabilityLookup<TKey> : IEnumerable<TKey>
    {
        private readonly object _sync = new object();

        private readonly List<int> _first = new List<int>();
        private readonly List<TKey> _second = new List<TKey>();

        private int _maximum = 0;
        private bool _frozen = false;

        protected internal PropabilityLookup()
        { }

        private void EnsureNotFrozen()
        {
            if (this._frozen)
            {
                throw new InvalidOperationException(); // Description
            }
        }

        protected void Add(TKey key, int propability)
        {
            Guard.EnsureNonnegative("propability", propability);

            lock (_sync)
            {
                EnsureNotFrozen();

                _first.Add(_maximum);
                _second.Add(key);

                checked { _maximum += propability; }
            }
        }

        protected void Freeze()
        {
            lock (_sync)
            {
                if (this._frozen)
                {
                    return;
                }

                this._first.Capacity = this._first.Count;
                this._second.Capacity = this._second.Count;

                this._frozen = true;
            }
        }

        protected TKey Get(Random r)
        {
            lock (_sync)
            {
                int result = this._first.BinarySearch(r.Next(this._maximum - 1));

                if (result < 0)
                {
                    result = ~result - 1;
                }

                return this._second[result];
            }
        }

        IEnumerator<TKey> IEnumerable<TKey>.GetEnumerator()
        {
            throw new NotSupportedException();
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new NotSupportedException();
        }
    }
}
