﻿using libdyn.tests.Utilities;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using Xunit;

namespace tests.Cli
{
    public static class DynConsoleIntegrationTests
    {
        [Theory(Skip = "Tests do not represent current state of libdyn.console")]
        [InlineData(@"",
            "",
            ":0:0: Error: No classes found.\n")]
        [InlineData(@"
class Main { }
",
            "",
            "")]
        public static void Console1(string standardInput, string expectedOutput, string expectedError)
        {
            Console2(null, standardInput, expectedOutput, expectedError);
        }
        [Theory(Skip = "Tests were not updated after complex changes in libdyn.console")]
        [InlineData(
            "--CONDITIONS\n--VERIFY",
            @"
class object { }
class numeric ensures false { method init() { @pred = null } }",
@"Assertion statement: true implies false
	Failure message: precondition of method init must imply weakest precondition
	Transformed expression : true implies false
	That statement is incorrect.
",
@"z3.out:0:0: Information: Verification condition satisfiable.
:0:0: Error: precondition of method init must imply weakest precondition
"
)]


        [InlineData(
            "--CONDITIONS\n--VERIFY",
            @"
class object { }
class numeric { method init() { @pred = null } }
class main { method run() requires false { } }",
@"Assertion statement: true implies true
	Failure message: precondition of method init must imply weakest precondition
	Transformed expression : true implies true
	That statement is correct.
Assertion statement: false implies true
	Failure message: precondition of method run must imply weakest precondition
	Transformed expression : false implies true
	That statement is correct.
",
@"z3.out:0:0: Information: Verification condition unsatisfiable.
z3.out:0:0: Information: Verification condition unsatisfiable.
")]
        [InlineData(
            "--CONDITIONS\n--VERIFY",
            @"
class object { }
class numeric { method init() { @pred = null } }
class main { method run() { } }",
@"Assertion statement: true implies true
	Failure message: precondition of method init must imply weakest precondition
	Transformed expression : true implies true
	That statement is correct.
Assertion statement: true implies true
	Failure message: precondition of method run must imply weakest precondition
	Transformed expression : true implies true
	That statement is correct.
",
@"z3.out:0:0: Information: Verification condition unsatisfiable.
z3.out:0:0: Information: Verification condition unsatisfiable.
")]
        [InlineData(
            "--CONDITIONS\n--VERIFY",
            @"
class object { }

class numeric {
    method init()
    {
        @pred = null
    }
}

class main {
    method run()
    {
        a = null;
        assert a == null;
        b = null;
        assert a == null and b == null
    }
}
",
            @"Assertion statement: true implies true
	Failure message: precondition of method init must imply weakest precondition
	Transformed expression : true implies true
	That statement is correct.
Assertion statement: a == null && b == null implies true
	Failure message: Assertion too weak.
	Transformed expression : a == null && b == null implies true
			Variable 'a' is of type 'Object'
			Variable 'b' is of type 'Object'
	That statement is correct.
Assertion statement: a == null implies true
	Failure message: Assertion too weak.
	Transformed expression : a == null implies true
			Variable 'a' is of type 'Object'
	That statement is correct.
Assertion statement: true implies true
	Failure message: precondition of method run must imply weakest precondition
	Transformed expression : true implies true
	That statement is correct.
", @"z3.out:0:0: Information: Verification condition unsatisfiable.
z3.out:0:0: Information: Verification condition unsatisfiable.
z3.out:0:0: Information: Verification condition unsatisfiable.
z3.out:0:0: Information: Verification condition unsatisfiable.
")]
        public static void Console2(string args, string standardInput, string expectedOutput, string expectedError)
        {
            StringBuilder @out = new StringBuilder(), @error = new StringBuilder();

            using (new ConsoleStreamsRedirector(standardInput, @out, @error))
            {
                var state = new Dyn.Console.Arguments();

                state.SolverPath = Constants.PATH_TO_Z3;

                Dyn.Console.Program.Run(state);
            }

            if(expectedOutput != null)
            {
                Assert.Equal(expectedOutput, @out.ToString(), false, true, true);
            }

            if (expectedError != null)
            {
                Assert.Equal(
                "Press ^D or ^Z to finish input.\n" + expectedError,
                @error.ToString(), false, true, true);
            }
        }

        public static readonly IEnumerable<object[]> EXAMPLES = SampleHelpers.Examples();

        [Theory]
        [MemberData("EXAMPLES")]
        public static void Examples(Sample input)
        {
            using (var reader = input.ToReader())
            {
                Console2("--CONDITIONS\n--VERIFY", reader.ReadToEnd(), null, null);
            }

        }
    }
}
