﻿using Dyn.Console;
using libdyn.tests.Utilities;
using System.Linq;
using System.Text;
using tests;
using Xunit;

namespace Dyn.Tests.Console
{
    using System;

    public static class ColoredFormatterUnitTests
    {
        private static readonly Lexer.SyntaxToken[] TEST_BENCH =
            Tokens.KINDS.SelectMany(Tokens.GetSamples).ToArray();
        private static readonly string TEST_BENCH_EXPECTED =
            string.Join("", TEST_BENCH.Select(t => t.ToSource()));

        [Fact]
        public static void Out()
        {
            var builderOut = new StringBuilder();
            var builderError = new StringBuilder();

            using (var cr = new ConsoleStreamsRedirector(null, builderOut, builderError))
            {
                ColoredFormatter.PrintTokenListHighlighted(TEST_BENCH, Console.Out);
            }

            Assert.Equal(TEST_BENCH_EXPECTED, builderOut.ToString());
            Assert.Equal(string.Empty, builderError.ToString());
        }

        [Fact]
        public static void Error()
        {
            var builderOut = new StringBuilder();
            var builderError = new StringBuilder();

            using (var cr = new ConsoleStreamsRedirector(null, builderOut, builderError))
            {
                ColoredFormatter.PrintTokenListHighlighted(TEST_BENCH, Console.Error);
            }

            Assert.Equal(string.Empty, builderOut.ToString(), false, true, false);
            Assert.Equal(TEST_BENCH_EXPECTED, builderError.ToString(), false, true, false);
        }

        [Fact]
        public static void Coverage()
        {
            ConsoleColor cc = ConsoleColor.Red;
            ColoredFormatter.ColorByKind(
                (Lexer.SyntaxKind)(short)-1,
                ref cc,
                ref cc);
            ColoredFormatter.ConsoleColorFromSeverity(
                (Diagnostics.DiagnosticSeverity) (short) -1,
                cc);
        }
    }
}
