﻿using System;
using System.Collections.Generic;
using System.Reflection;
using Dyn.Console.CommandLine;
using Xunit;

namespace Dyn.Tests.Console.CommandLine
{
    public static class CommandLineParserUnitTests
    {
        private class PositiveOptions
        {
            private IEnumerator<string> enumerator;

            public void SetExpected(IEnumerable<string> expected)
            {
                this.enumerator = expected.GetEnumerator();
            }


            [Option(IsArgument = true)]
            [Option(ShortName = 'a', LongName = "b")]
            public void Method(string argument)
            {
                Assert.NotNull(this.enumerator);
                Assert.True(this.enumerator.MoveNext());
                Assert.Equal(this.enumerator.Current, argument);
            }
        }

        private class NegativeOptions1
        {
            [Option(Description = "1", ShortName = 'a')]
            [Option(Description = "2", ShortName = 'a')]
            public string Value;
        }

        private class NegativeOptions2
        {
            [Option(Description = "1", LongName = "b")]
            [Option(Description = "2", LongName = "b")]
            public string Value;
        }

        private class NegativeOptions3
        {
            [Option(Description = "1", IsArgument = true)]
            [Option(Description = "2", IsArgument = true)]
            public string Value;
        }

        [Theory]
        [InlineData("|C|A|B", "|C|-a|A|--b:B")]
        [InlineData("|D|B|A", "|D|--b=B|-aA")]
        [InlineData("|B|A|E", "|--b=B|-aA|E")]
        [InlineData("|A|B|F", "|-a|A|--b:B|F")]
        public static void Positive(string expected, string input)
        {
            var instance = new PositiveOptions();
            instance.SetExpected(
                CommandLineLexerUnitTests.SplitWithFirstLineContent(expected));
            CommandLineParser<PositiveOptions>.Parse(
                CommandLineLexerUnitTests.SplitWithFirstLineContent(input),
                instance);
        }

        [Fact]
        [System.Diagnostics.DebuggerHidden]
        public static void Negative()
        {
            string[] e = CommandLineLexerUnitTests.SplitWithFirstLineContent(
                "|-a:b|--C=d|E");

            Assert.NotNull(new NegativeOptions1().Value = new NegativeOptions2().Value = new NegativeOptions3().Value = string.Empty);

            TypeInitializationException ex;
            ex = Assert.Throws<TypeInitializationException>(
                () => CommandLineParser<NegativeOptions1>.Parse(e, new NegativeOptions1()));
            Assert.IsType<OptionDefinitionException>(ex.InnerException);
            ex = Assert.Throws<TypeInitializationException>(
                () => CommandLineParser<NegativeOptions2>.Parse(e, new NegativeOptions2()));
            Assert.IsType<OptionDefinitionException>(ex.InnerException);
            ex = Assert.Throws<TypeInitializationException>(
                () => CommandLineParser<NegativeOptions3>.Parse(e, new NegativeOptions3()));
            Assert.IsType<OptionDefinitionException>(ex.InnerException);
            Assert.ThrowsAny<ArgumentParserException>(() => Positive("", "|-c|"));
        }
    }
}