﻿using Dyn.Console;
using Dyn.Diagnostics;
using Dyn.Lexer;
using libdyn.tests.Utilities;
using System.Linq;
using System.Text;
using tests;
using tests.Ast;
using Xunit;

namespace Dyn.Tests.Console
{
    using System;
    using Core.Diagnostics;
    static class ConsoleDiagnosticHandlerUnitTests
    {
        [Fact]
        public static void Simple()
        {
            var cdh = new ConsoleDiagnosticHandler(new SyntaxToken[0]);

            var r = new Random(0);

            using (var slim = new ConsoleStreamsRedirector(null, null, null))
            {
                IDiagnosticUnitTests.FuzzingEachSeverity(cdh, r);
                IDiagnosticUnitTests.FuzzingBatch(cdh, 100, r);

                cdh.Invoke(
                    null,
                    DiagnosticSeverity.None,
                    new Span(@"$@\/\/ \n", 0, 0, 0, 0),
                    "Some message");

                cdh.Invoke(
                    null,
                    DiagnosticSeverity.None,
                    default(Span),
                    "Some message");

                Console.ResetColor();
            }
        }

        public static void AssertConsoleDiagnosticInputOutput(string expected, string input, DiagnosticInfo diagnostic)
        {
            var tokens = input.Tokenize().ToList();

            var builder = new StringBuilder();

            using (var slim = new ConsoleStreamsRedirector(null, builder, builder))
            {
                var cdh = new ConsoleDiagnosticHandler(tokens);
                cdh.DiagnosticInfoListener(null, new DiagnosticEventArgs(diagnostic));
                Console.ResetColor();
            }

            Assert.Equal(expected, builder.ToString(), false, true, false);
        }



        [Theory]
        [InlineData(@":0:0: $1: $2


", @"", 0, 0, 0, 0)]
        [InlineData(@":0:0: $1: $2
1 2 3
^
", @"1 2 3", 0, 0, 0, 1)]
        [InlineData(@":0:2: $1: $2
1 2 3
  ^
", @"1 2 3", 2, 0, 2, 1)]
        [InlineData(@":0:4: $1: $2
1 2 3
    ^
", @"1 2 3", 4, 0, 4, 1)]
        [InlineData(@":0:2: $1: $2
A simple single line.
  ^^^^^^
", @"A simple single line.", 2, 0, 2, 6)]
        public static void SingleLine(string expected, string input, int pos, int line, int column, int length)
        {
            var r = new Random(Tuple.Create(expected, input, pos, line, column, length).GetHashCode());

            var diagnostic = new DiagnosticInfo(
                IDiagnosticUnitTests.Severities.Random(r),
                new Span(string.Empty, pos, line, column, length),
                DynFuzzer.FuzzyString(r, 0));

            Func<string, string> fmt = e => e.Replace("$1", diagnostic.Level.ToString()).Replace("$2", diagnostic.Message);

            AssertConsoleDiagnosticInputOutput(
                fmt(expected),
                input.ToString(),
                diagnostic);
        }

        [Fact(Skip = "Not verified here. It uses SyntaxTokenExtensions.ExtractSpan. This method will have its own tests.")]
        public static void MultiLine()
        {
            throw new NotImplementedException();
        }
    }
}