﻿using Dyn.Console;
using Dyn.Diagnostics;
using Dyn.Lexer;
using System;
using System.Collections.Generic;
using System.Linq;
using tests;
using Xunit;
using Dyn.Tests.Core.Diagnostics;

namespace Dyn.Tests.Console
{
    public static class DiagnosticSummaryUnitTests
    {
        private static readonly DiagnosticSeverity[] SEVERITIES = Utils.GetAllValuesEnum<DiagnosticSeverity>();

        [Theory]
        [InlineData(0, 0, 0, 0)]
        [InlineData(10, 0, 0, 0)]
        [InlineData(0, 20, 0, 0)]
        [InlineData(0, 0, 30, 0)]
        [InlineData(0, 0, 0, 40)]
        [InlineData(10, 0, 0, 0)]
        [InlineData(10, 20, 0, 0)]
        [InlineData(10, 20, 30, 0)]
        [InlineData(10, 20, 30, 40)]
        public static void Simple(int none, int info, int warning, int error)
        {
            var r = new Random(1);

            var testee = new DiagnosticSummary();

            Assert.Equal(testee[DiagnosticSeverity.None], 0);
            Assert.Equal(testee[DiagnosticSeverity.Information], 0);
            Assert.Equal(testee[DiagnosticSeverity.Warning], 0);
            Assert.Equal(testee[DiagnosticSeverity.Error], 0);

            IDiagnosticUnitTests.AssertPassThrough(testee,
                    Dyn.Utilities.AsArray(
                        Enumerable.Repeat(DiagnosticSeverity.None, none),
                        Enumerable.Repeat(DiagnosticSeverity.Information, info),
                        Enumerable.Repeat(DiagnosticSeverity.Warning, warning),
                        Enumerable.Repeat(DiagnosticSeverity.Error, error)).
                    SelectMany(e => e).Select(e => IDiagnosticUnitTests.FuzzerDiagnosticInfo(e, r)));

            Assert.Equal(testee[DiagnosticSeverity.None], none);
            Assert.Equal(testee[DiagnosticSeverity.Information], info);
            Assert.Equal(testee[DiagnosticSeverity.Warning], warning);
            Assert.Equal(testee[DiagnosticSeverity.Error], error);

            testee.Invoke(null, DiagnosticSeverity.None, Span.Empty, string.Empty);
        }

        [Fact]
        public static void Counting1()
        {
            var testee = new DiagnosticSummary();

            foreach (var e in SEVERITIES)
            {
                Assert.Equal(testee[e], 0);
            }

            IDiagnosticUnitTests.FuzzingEachSeverity(testee, new Random(1));

            foreach (var e in SEVERITIES)
            {
                Assert.Equal(testee[e], 1);
            }

        }


        [Theory]
        [InlineData(0, 100)]
        [InlineData(1, 200)]
        [InlineData(2, 300)]
        public static void Counting2(int seed, int count)
        {
            var r = new Random(seed);

            var testee = new DiagnosticSummary();

            var counter = SEVERITIES.ToDictionary(e => e, e => 0);

            foreach (var e in SEVERITIES)
            {
                Assert.Equal(testee[e], counter[e]);
            }

            IDiagnosticUnitTests.AssertPassThrough(
                testee,
                Enumerable.Range(0, count).Select(e =>
                {
                    var current = SEVERITIES.Random(r);
                    counter[current] = counter[current] + 1;

                    return IDiagnosticUnitTests.FuzzerDiagnosticInfo(current, r);
                }));

            foreach (var e in SEVERITIES)
            {
                Assert.Equal(testee[e], counter[e]);
            }
        }
    }
}
