﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Dyn.Tests.Console.CommandLine;
using libdyn.tests.Utilities;
using tests;
using Xunit;

namespace Dyn.Tests.Console
{
    public static class ProgramUnitTests
    {
        const string HELP = @"
libdyn.console [-i] [-achs] [-?] -z path [file...]
libdyn.console formally verifies Dyn programs.

Commands:
-i      --interactive                  reads stdin for user input
                                       (automatically activated if no 
                                       files are provided)
-z      --solver           PATH        executes a SMTLIB2 solver with PATH
-s      --show-summary                 prints a summary
-h      --show-highlighted             prints highlighted source files
-v      --verify                       verifies the method contracts
-c      --show-conditions              prints verification conditions
-a      --show-ast                     prints abstract syntax trees 
-?      --help                         prints this message
";

        [Theory]
        [InlineData("",
            ":0:0: Error: No classes found.\r\n",
            "",
            "|-h")]
        [InlineData(HELP,
            "",
            @"",
            "|-?")]
        [InlineData(HELP,
            "",
            @"",
            "|-ü")]
//        [InlineData(@"class object
//{}


//class nullclass
//{}


//class numeric
//   ensures false
//{
//   method init()
//   {
//      @pred = null
//   }
//}
//Verification condition summary:
//	       1 found.
//	       0 successfully verified.
//	       0 skipped.
//	       1 failed to verify.
//Diagnostic Summary:
//	       1 other(s)
//	       1 info(s)
//	       0 warning(s)
//	       1 error(s)
//",
//            @"Press ^D or ^Z to finish input.
//Verification request 1:
//	Assertion: true implies false
//	Description: precondition of method numeric.init must imply weakest precondition
//	Ignored: False
//	Transformed Assertion : true implies false
//z3.in:0:0: None: (set-option :produce-unsat-cores true)
//(set-option :produce-models true)
//;(set-option :produce-proofs true)

//; --------------- Basic Definitions -------------------

//(declare-sort Obj 0)

//(declare-const Null Obj)

//; --------------- Classes -------------------

//(declare-sort Class 0)

//(declare-fun class(Obj) Class)
//(declare-const class_object Class)
//(declare-const class_nullclass Class)
//(declare-const class_numeric Class)
//(assert (distinct class_object class_nullclass class_numeric))

//(assert (= (class Null) class_nullclass))

//; --------------- Instance variables -------------------

//(declare-sort IVar 0)

//(declare-const iv_pred IVar)
//(assert (distinct iv_pred))

//(declare-fun inst_ref(Obj IVar) Obj)
//(declare-const this Obj)

//; --------------- Predicates -------------------------------



//; --------------- Contextual Information -------------------

//(declare-const lv_true Bool)
//(declare-const lv_false Bool)

//(assert (= lv_true true))
//(assert (= lv_false false))



//; --------------- Contextual Information -------------------
//(declare-const pv_self Obj)
//(assert (= (class pv_self) class_numeric))

//; --------------- Verification Condition -------------------



//(assert (not (=> lv_true lv_false)))

//; --------------- Verify -------------------
//(check-sat)

//z3.out:0:0: Information: Verification condition invalid.
//:0:0: Error: precondition of method numeric.init must imply weakest precondition
//",
//            @"
//class object { }
//class nullclass { }
//class numeric ensures false { method init() { @pred = null } }",
//    "|-z $Z3|-civas"
//            )]
        public static void AssertSimple(string expectedOut, string expectedError, string stdIn, string args)
        {
            var builderOut = new StringBuilder();
            var builderError = new StringBuilder();

            var argsArray = CommandLineLexerUnitTests.SplitWithFirstLineContent(
                args.Replace("$Z3", Constants.PATH_TO_Z3));

            using (var cr = new ConsoleStreamsRedirector(stdIn, builderOut, builderError))
            {
                Dyn.Console.Program.Main(argsArray);
            }

            Assert.Equal(expectedOut, builderOut.ToString(), false, true, false);
            Assert.Equal(expectedError, builderError.ToString(), false, true, false);
        }
    }
}
