﻿using System;
using System.Collections.Generic;
using System.Linq;
using Dyn.Diagnostics;
using Dyn.Lexer;
using tests;
using Xunit;

namespace Dyn.Tests.Core.Diagnostics
{
    public static class IDiagnosticUnitTests
    {
        public static readonly DiagnosticSeverity[] Severities = Utils.GetAllValuesEnum<DiagnosticSeverity>();

        public static DiagnosticInfo FuzzerDiagnosticInfo(Random r)
        {
            return FuzzerDiagnosticInfo(Severities.Random(r), r);
        }

        public static DiagnosticInfo FuzzerDiagnosticInfo(DiagnosticSeverity severity, Random r)
        {
            return new DiagnosticInfo(
                severity,
                new Span(
                    tests.Ast.DynFuzzer.FuzzyString(r, 0),
                    2 + r.Next(short.MaxValue),
                    1 + r.Next(short.MaxValue),
                    1 + r.Next(short.MaxValue),
                    r.Next(short.MaxValue)),
                tests.Ast.DynFuzzer.FuzzyString(r, 0));
        }

        public static void AssertPassThrough(
            IDiagnostic testee,
            IEnumerable<DiagnosticInfo> expected)
        {
            DiagnosticInfo actual = null;

            EventHandler<DiagnosticEventArgs> listener = (_, e) =>
            {
                actual = e.Info;
            };

            testee.DiagnosticInfo += listener;

            foreach (var element in expected)
            {
                testee.DiagnosticInfoListener(null, new DiagnosticEventArgs(element));
                Assert.Equal(element, actual);
                actual = null;
            }

            testee.DiagnosticInfo -= listener;
        }


        public static void FuzzingEachSeverity(IDiagnostic testee, Random r)
        {
            AssertPassThrough(
                testee,
                Severities.Select(e => FuzzerDiagnosticInfo(e, r)));
        }

        public static void FuzzingBatch(IDiagnostic testee, int count, Random r)
        {
            AssertPassThrough(
                testee,
                Enumerable.Range(0, count).Select(e => FuzzerDiagnosticInfo(Severities.Random(r), r)));
        }
    }
}
