﻿using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;

namespace Dyn.Tests.Execution
{
    using System;

    public static class ConsoleRunner
    {
        public static void WriteSummaryLine(ConsoleColor fg, int count, string text)
        {
            var t = Console.ForegroundColor;
            Console.ForegroundColor = fg;
            Console.Write(count.ToString().PadLeft(8));
            Console.ForegroundColor = t;
            Console.WriteLine(text);
        }

        private static Result TestWrapper(Test test)
        {
            var testResult = test.Test();

            if (testResult.Fails + testResult.Errors > 0)
            {
                Console.Out.WriteLine("{0}:{1}",
                    test.Method.DeclaringType.FullName,
                    test.Method.Name);
            }

            return testResult;
        }

        public static Result TestConsole(this IEnumerable<Test> source)
        {
            var result = source.AsParallel().Select(TestWrapper).Aggregate((a, b) => a + b);

            WriteSummaryLine(Console.ForegroundColor, result.Total, " tests found.");
            WriteSummaryLine(ConsoleColor.Green, result.Passes, " successfully verified.");
            WriteSummaryLine(ConsoleColor.Yellow, result.Skips, " skipped.");
            WriteSummaryLine(ConsoleColor.Red, result.Fails, " failed to verify.");
            WriteSummaryLine(ConsoleColor.DarkRed, result.Errors, " execution failures.");
            Console.Out.Flush();

            return result;
        }
    }
}
