﻿using System;

namespace Dyn.Tests.Execution
{
    public struct Result
    {
        private readonly int _passes, _skips, _fails, _errors;
        private static readonly Result
            _pass = new Result(1, 0, 0, 0),
            _skip = new Result(0, 1, 0, 0),
            _fail = new Result(0, 0, 1, 0),
            _error = new Result(0, 0, 0, 1),
            _empty = new Result(0, 0, 0, 0);

        public Result(int passes, int skips, int fails, int errors)
        {
            Guard.EnsureNonnegative("passes", passes);
            Guard.EnsureNonnegative("skips", skips);
            Guard.EnsureNonnegative("fails", fails);
            Guard.EnsureNonnegative("errors", errors);

            this._passes = passes;
            this._skips = skips;
            this._fails = fails;
            this._errors = errors;
        }

        public static Result Fail { get { return _fail; } }
        public static Result Pass { get { return _pass; } }
        public static Result Skip { get { return _skip; } }
        public static Result Error { get { return _error; } }
        public static Result Empty { get { return _empty; } }

        public static Result operator +(Result a, Result b)
        {
            return new Result(
                a.Passes + b.Passes,
                a.Skips + b.Skips,
                a.Fails + b.Fails,
                a.Errors + b.Errors);
        }

        public Result Scale(int scalar)
        {
            Guard.EnsureNonnegative("scalar", scalar);

            return new Result(
                this.Passes * scalar,
                this.Skips * scalar,
                this.Fails * scalar,
                this.Errors * scalar);
        }


        public static Result operator *(Result a, int scalar)
        {
            return a.Scale(scalar);
        }

        public static Result operator *(int scalar, Result a)
        {
            return a.Scale(scalar);
        }

        public int Passes { get { return this._passes; } }
        public int Skips { get { return this._skips; } }
        public int Fails { get { return this._fails; } }
        public int Errors { get { return this._errors; } }
        public int Total
        {
            get
            {
                return this._passes +
                    this._skips +
                    this._fails +
                    this._errors;
            }
        }

    }
}
