﻿using System;
using System.Collections.Generic;

namespace Dyn.Tests.Execution
{
    public static class Runner
    {
        public static Result Test(this IEnumerable<Test> source)
        {
            Guard.EnsureNotNull("source", source);

            var result = Result.Empty;

            foreach (var element in source)
            {
                result += element.Test();
            }

            return result;
        }

        public static Result Test(this Test test)
        {
            if (test.Action == null)
            {
                throw new ArgumentNullException("Element is not valid.");
            }

            if(!string.IsNullOrEmpty(test.Attribute.Skip))
            {
                return Result.Skip;
            }

            try
            {
                test.Action();
                return Result.Pass;
            }
            catch (System.Reflection.TargetInvocationException ex)
            {
                var inner = ex.InnerException;

                if(inner is Xunit.Sdk.XunitException)
                {
                    System.Console.Error.WriteLine(inner.ToString());
                    return Result.Fail;
                }
                else
                {
                    return Result.Error;
                }
            }
            // super errors
            catch (System.Reflection.TargetException)
            {
                return Result.Error;
            }
            catch (System.ArgumentException)
            {
                return Result.Error;
            }
            catch (System.Reflection.TargetParameterCountException)
            {
                return Result.Error;
            }
            catch (System.MethodAccessException)
            {
                return Result.Error;
            }
            catch (System.InvalidOperationException)
            {
                return Result.Error;
            }
            catch (System.NotSupportedException)
            {
                return Result.Error;
            }
        }
    }
}
