﻿using Dyn;
using System;
using System.Collections.Generic;

namespace tests.Lexer
{
    internal static class SyntaxTokenAnalyser
    {
        private class ArrayComparer<T> : IEqualityComparer<T[]>
        {
            private readonly IEqualityComparer<T> _comparer;

            public ArrayComparer(IEqualityComparer<T> comparer)
            {
                Guard.EnsureNotNull("comparer", comparer);

                this._comparer = comparer;
            }

            public bool Equals(T[] x, T[] y)
            {
                if (x.Length != y.Length)
                {
                    return false;
                }

                for (int i = 0; i < x.Length; ++i)
                {
                    if (!_comparer.Equals(x[i], y[i]))
                    {
                        return false;
                    }
                }

                return true;
            }

            public int GetHashCode(T[] obj)
            {
                unchecked
                {
                    const int SEED = 1;
                    const int PRIME = 31;

                    int result = SEED;

                    for (int i = 0; i < obj.Length; ++i)
                    {
                        result = result * PRIME ^ _comparer.GetHashCode(obj[i]);
                    }

                    return result;
                }
            }
        }

        public static IDictionary<T[], int> Analyse<T>(IEnumerable<T> source, int rank, IEqualityComparer<T> comparer = null)
        {
            Guard.EnsureNotNull("source", source);
            IDictionary<T[], int> result = new Dictionary<T[], int>(
                new ArrayComparer<T>(comparer ?? EqualityComparer<T>.Default));

            if (rank < 1)
            {
                throw new ArgumentOutOfRangeException("rank");
            }

            T[] array = new T[rank];

            int i = 0;

            var enumerator = source.GetEnumerator();

            while (i < rank && enumerator.MoveNext())
            {
                array[i] = enumerator.Current;
                ++i;
            }

            if (i < rank)
            {
                return result;
            }

            while (true)
            {
                int value;
                if (result.TryGetValue(array, out value))
                {
                    ++value;
                }
                else
                {
                    value = 1;
                }

                result[(T[])array.Clone()] = value;

                if (!enumerator.MoveNext())
                {
                    break;
                }

                Array.Copy(array, 1, array, 0, rank - 1);

                array[rank - 1] = enumerator.Current;
            }

            return result;
        }
    }
}
