﻿using System;
using System.IO;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading;
using Xunit;

namespace libdyn.tests.Utilities
{
    public sealed class ConsoleStreamsRedirector : IDisposable
    {
        //private static void CurrentThreadInfo(string prefix)
        //{
        //    var ct = Thread.CurrentThread;

        //    System.Diagnostics.Debug.WriteLine(
        //        "{0}\t {1}: {2}",
        //        prefix,
        //        ct.ManagedThreadId,
        //        ct.Name);
        //}

        private readonly TextReader _oldIn;
        private readonly TextWriter _oldOut;
        private readonly TextWriter _oldErr;
        private static readonly object syncLock = new object();

        public ConsoleStreamsRedirector(string @in, StringBuilder @out, StringBuilder @error)
        {
            Monitor.Enter(syncLock);
            //CurrentThreadInfo(">");

            this._oldIn = Console.In;
            this._oldOut = Console.Out;
            this._oldErr = Console.Error;

            Console.SetIn(
                @in == null ?TextReader.Null : new StringReader(@in));
            Console.SetOut(
                @out == null ? TextWriter.Null : new StringWriter(@out));
            Console.SetError(
                @error == null ? TextWriter.Null : new StringWriter(@error));
        }

        ~ConsoleStreamsRedirector()
        {
            this.Dispose(false);
        }

        private bool disposed = false;

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        private void Dispose(bool disposing)
        {
            if(disposed)
            {
                return;
            }

            Console.SetIn(this._oldIn);
            Console.SetOut(this._oldOut);
            Console.SetError(this._oldErr);

            disposed = true;
            //CurrentThreadInfo("<");
            Monitor.Exit(syncLock);
        }
    }
}
