﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Dyn.Tests.Utilities
{
    public sealed class ArrayEqualityComparer<T> : IEqualityComparer<T[]>
    {
        private readonly IEqualityComparer<T> _equalityComparer;

        public ArrayEqualityComparer(IEqualityComparer<T> equalityComparer)
        {
            Guard.EnsureNotNull("equalityComparer", equalityComparer);

            this._equalityComparer = equalityComparer;
        }

        public bool Equals(T[] x, T[] y)
        {
            if (x == null)
            {
                return y == null;
            }

            if (y == null)
            {
                return false;
            }

            if (object.ReferenceEquals(x, y))
            {
                return true;
            }

            if (x.Length != y.Length)
            {
                return false;
            }

            for (int i = 0, max = x.Length; i < max; ++i)
            {
                if (!_equalityComparer.Equals(x[i], y[i]))
                {
                    return false;
                }
            }
            return true;
        }

        public int GetHashCode(T[] obj)
        {
            if (obj == null)
            {
                return 0;
            }

            int hash = obj.Length;

            unchecked
            {
                for (int i = 0, max = obj.Length; i < max; ++i)
                {
                    hash = hash << 5 + hash ^ this._equalityComparer.GetHashCode(obj[i]);
                }

                return hash;
            }
        }
    }
}
