﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;

namespace tests
{
    internal static class Utils
    {
        public static T[] GetAllValuesEnum<T>()
        {
            return Enum.GetValues(typeof(T)).Cast<T>().ToArray();
        }

        public static T Random<T>(this IList<T> source, Random rnd)
        {
            return source[rnd.Next(0, source.Count - 1)];
        }

        public static void Shuffle<T>(this IList<T> source, Random r)
        {
            int n = source.Count;
            while (n > 1)
            {
                n--;
                int k = r.Next(n);
                T value = source[k];
                source[k] = source[n];
                source[n] = value;
            }
        }

        public static string ToGenericTypeString(this Type t)
        {
            if (!t.GetTypeInfo().IsGenericType)
                return t.Name;
            string genericTypeName = t.GetGenericTypeDefinition().Name;
            genericTypeName = genericTypeName.Substring(0,
                genericTypeName.IndexOf('`'));
            string genericArgs = string.Join(",",
                t.GetTypeInfo().GenericTypeArguments
                    .Select(ta => ToGenericTypeString(ta)).ToArray());
            return genericTypeName + "<" + genericArgs + ">";
        }

        public static T As<T>(this object source)
        {
            return (T)source;
        }

        public static IEnumerable<object[]> FiXunit<T>(this IEnumerable<T> source)
        {
            return source.Select(e => new object[1] { e });
        }
    }
}
