﻿using Dyn;
using Dyn.Ast;
using Dyn.Diagnostics;
using Dyn.VerficationConditions;
using System;
using Xunit;

namespace tests.Wpc
{
    public static class AtiUnitTests
    {
        [Theory]
        [InlineData(null, "1 == 1")]
        [InlineData(null, "1 == 2")]
        [InlineData(null, "null == null")]
        [InlineData(null, "true == false")]
        [InlineData(null, "2 + 5 == 5")]
        [InlineData(null, "(1 == 2) == 3 < 4")]
        public static void Positive(string expected, string actual)
        {
            if (expected == null)
            {
                expected = actual;
            }

            var expression = actual.Tokenize().Parse(true, p => p.ParseAssertionExpression());

            var result = DynAssertionTypeInference.Run(Magic.ThrowOnDiagnostic, (e) => null, expression, string.Empty).Expression;

            Assert.NotEqual(AstExtensions.Assertions.False, result, AstExtensions.IgnoreLocation);
            Assert.Equal(expected, result.ToSource());
            AssertExtensions.SyntaxEqual(expected, result.ToSource());
        }

        [Theory]
        [InlineData("1 == null")]
        public static void Negative(string actual)
        {
            var counter = 0;

            EventHandler<DiagnosticEventArgs> callback = (sender, args) => counter++;

            var expression = actual.Tokenize().Parse(true, p => p.ParseAssertionExpression());

            var falseExpression = DynAssertionTypeInference.Run(callback, (e) => null, expression, string.Empty).Expression;

            Assert.Equal(AstExtensions.Assertions.False, falseExpression, AstExtensions.IgnoreLocation);
            Assert.True(counter > 0);
        }
    }
}
