﻿using Dyn.Ast;
using Dyn.VerficationConditions;
using Xunit;

namespace tests.Wpc
{
    public static class VcUnitTests
    {
        [Theory]
        [InlineData("Null", "null")]
        [InlineData("lv_true", "true")]
        [InlineData("lv_false", "false")]
        [InlineData("pv_a", "a")]
        [InlineData("1", "1")]
        [InlineData("lv_a", "'a")]
        [InlineData("(not pv_a)", "!a")]
        [InlineData("(+ lv_a pv_b)", "'a + b")]
        [InlineData("(- lv_a pv_b)", "'a - b")]
        [InlineData("(* lv_a pv_b)", "'a * b")]
        [InlineData("(div pv_a pv_b)", "a / b")]
        [InlineData("(mod pv_a pv_b)", "a mod b")]
        [InlineData("(< lv_a pv_b)", "'a < b")]
        [InlineData("(<= lv_a pv_b)", "'a <= b")]
        [InlineData("(> lv_a pv_b)", "'a > b")]
        [InlineData("(>= lv_a pv_b)", "'a >= b")]
        [InlineData("(= lv_a pv_b)", "'a == b")]
        [InlineData("(and lv_a pv_b)", "'a && b")]
        [InlineData("(or lv_a pv_b)", "'a || b")]
        [InlineData("(inst_ref pv_self iv_a)", "self.@a")]
        [InlineData("(ite pv_a pv_b pv_c)", "if a then b else c end")]
        [InlineData("(first)", "first()")]
        [InlineData("(second pv_a)", "second(a)")]
        [InlineData("(third pv_a pv_b)", "third(a, b)")]
        [InlineData("(forth pv_a pv_b pv_c)", "forth(a, b, c)")]
        [InlineData("(or (= (class pv_a) class_b) (= (class pv_a) class_c) (= (class pv_a) class_d))", "a in { b, c, d }")]
        [InlineData("(forall ( (lv_a b)) pv_c)", "forall 'a in b: c")]
        [InlineData("(exists ( (lv_a b)) pv_c)", "exists 'a in b: c")]
        public static void Simple(string expected, string actual)
        {
            Assert.Equal(expected, DynAssertionSmtLibFormatter.Visit(
                actual.Tokenize().Parse(true, p => p.ParseAssertionExpression())).ToString());
        }

        [Theory]
        [InlineData("'a + 'a == 1", "'a + 'a == 1")]
        [InlineData("false implies true", "false implies true")]
        [InlineData("result == null", "result == null")]
        public static void SimpleTransform(string expected, string input)
        {
            AssertExtensions.SyntaxEqual(
                expected,
                DynAssertionTypeInference.Run(
                    Magic.ThrowOnDiagnostic,
                    e => null,
                    input.Tokenize().Parse(
                        true,
                        p => p.ParseAssertionExpression()), string.Empty).Expression.ToSource());
        }
    }
}
