﻿using Dyn.Ast;
using Dyn.VerficationConditions;
using System.Collections.Generic;
using System.Linq;
using Xunit;

namespace tests.Wpc
{
    public static class WpcUnitTests
    {
        public static string SOURCE = @"
class dyn_lang_builtin
{
    method a(b, c) requires b == c ensures result == null
    { }

    method d(e) requires e == null ensures result == 1
    { }
}
class x2
{
    method a(b, c) requires b == null && 1 > 2 ensures result == null
    { }

    method d(e, f, g) requires e == (f == g) ensures result == 3
    { }
}
class test1
{
    method init(b) requires b == null
    { }
}";

        public static DynProgramClassNode[] SOURCE_CLASSES = SOURCE.Tokenize().ParseAll(true, p => p.ParseClass()).ToArray();

        [Theory]
        [InlineData(@"{result == null}

{result == null}")]
        [InlineData(@"{result == null}

{result == null}")]
        [InlineData(@"{x < 5}

{x < 5}")]
        [InlineData(@"{null == null}
x = null
{x == null}")]
        [InlineData(@"{if self == self then null else @x end == null}
@x = null
{@x == null}")]
        [InlineData(@"{if self == self then null else @x end == null}
@x = null
{self.@x == null}")]
        [InlineData(@"{@y == null}
@x = null
{@y == null}")]
        [InlineData(@"{@x == null}
@y = null
{@x == null}")]
        [InlineData(@"{self.@y == null}
@x = null
{self.@y == null}")]
        [InlineData(@"{
self in {dyn_lang_builtin} && (null == null && (
    forall '_a2 : object. forall '_a1 : object. forall '_a0 : object.
    forall '_result : object. '_result == null implies self.@y == null)) || 
(
    self in {x2} && (null == null && 1 > 2 && (
    forall '_a2 : object. forall '_a1 : object. forall '_a0 : object.
    forall '_result : object. '_result == null implies self.@y == null)))}
self.a(null, null)
{self.@y == null}")]
        [InlineData(@"{}
while a invariant true do null done
{}")]
        [InlineData(@"{false}
new test123(null) # no constructor found
{}")]
        [InlineData(@"{null == null}
new test1(null) # no constructor found
{}")]
        public static void Postive(string hoareSource)
        {
            var hoare = hoareSource.Tokenize().Parse(true, p => p.ParseHoare());
            Assert.NotNull(hoare);

            List<VerificationRequest> requests = new List<VerificationRequest>();

            var vcg = new VerficationConditionGenerator(SOURCE_CLASSES);
            vcg.DiagnosticInfo += Magic.ThrowOnDiagnostic;
            var wpc = new WeakestPreconditionCalculus(vcg, requests.Add);
            wpc.DiagnosticInfo += Magic.ThrowOnDiagnostic;
            var pre = wpc.Wp(hoare.Statements, hoare.PostCondition);

            Assert.Equal(hoare.PreCondition.ToSource(), pre.ToSource());
            AssertExtensions.SyntaxEqual(hoare.PreCondition.ToSource(), pre.ToSource());
        }
    }
}
