﻿using Dyn.Ast;
using Dyn.VerficationConditions;
using System.Collections.Generic;
using Xunit;

namespace tests.Wpc
{
    public static class Z3IntegrationTests
    {
        private static string CONTEXT_SOURCE = @"class numeric { method test() { @pred; } }";

        private static IEnumerable<DynProgramClassNode> CONTEXT = CONTEXT_SOURCE.Tokenize().ParseAll(true, p => p.ParseClass());

        [Theory(Skip = "Tests were not updated after complex changes in Z3")]
        [InlineData(Z3Adapter.Z3Result.Unsatisfiable, "1 == 1")]
        [InlineData(Z3Adapter.Z3Result.Satisfiable, "1 == 2")]
        [InlineData(Z3Adapter.Z3Result.Unsatisfiable, "(1 + 1) == 2")]
        [InlineData(Z3Adapter.Z3Result.Satisfiable, "(1 + 1) == 3")]
        [InlineData(Z3Adapter.Z3Result.Unsatisfiable, "('x > 0) implies ('x + 'x >= 'x)")]
        [InlineData(Z3Adapter.Z3Result.Unsatisfiable, "('x < 5 implies 'x < 10)")]
        [InlineData(Z3Adapter.Z3Result.Satisfiable, "('x < 10 implies 'x < 5)")]
        public static void Simple(Z3Adapter.Z3Result expected, string input)
        {
            AssertSimple(CONTEXT, input.Tokenize().Parse(true, p => p.ParseAssertionExpression()), expected);
        }

        public static void AssertSimple(
            IEnumerable<DynProgramClassNode> context,
            DynAssertionExpressionNode expression,
            Z3Adapter.Z3Result expectedResult)
        {
            var z3 = new Z3Adapter(
                 Constants.PATH_TO_Z3,
                context, string.Empty);

            var result = DynAssertionTypeInference.Run(Magic.ThrowOnError, e => null, expression, string.Empty);

            z3.DiagnosticInfo += Magic.ThrowOnError;

            Assert.True(expectedResult == z3.Run(result), "Input: " + result.Expression.ToSource());
        }
    }
}
